#! /usr/env/python
"""
Landlab's Continuous-Time Stochastic (CTS) cellular automata modeling package.

Overview
--------

A CellLab CTS model implements a particular type of cellular
automaton (CA): a continuous-time stochastic CA. The approach is based on that
of Narteau et al. (2002, 2009) and Rozier and Narteau (2014). Like a normal
CA, the domain consists of a lattice of cells, each of which has a discrete
state. Unlike a conventional CA, the updating process is stochastic, and takes
place in continuous rather than discrete time. Any given pair (or "doublet")
of adjacent cell states has a certain specified probability of transition to a
different pair of states. The transition probability is given in the form of an
average *transition rate*, :math:`\lambda` (with dimensions of 1/T); the actual
time of transition is a random variable drawn from an exponential probability
distribution with mean :math:`1/\lambda`.

Subclasses
----------

Landlab provides for several different lattice and connection types:

-  RasterCTS: regular raster grid with transitions between horizontal and
   vertical cell pairs
-  OrientedRasterCTS: like a RasterLCA, but different transition rates can
   be assigned to vertical and horizontal pairs. This property of
   orientation can be used, for example, to implement rules representing
   gravitational attraction, or flow of a fluid with a particular
   direction.
-  RasterD8CTS: like a RasterLCA, but includes diagonal as well as vertical
   and horizontal cell pairs.
-  OrientedRasterD8CTS: as above but orientation also matters.
-  HexCTS: hexagonal grid
-  OrientedHexCTS: hexagonal grid, with transition rates allowed to vary
   according to orientation.

Encoding of "states"
--------------------
As in any traditional cellular automaton model, a LandlabCellularAutomaton
contains a grid of cells ("nodes" in Landlab parlance), each of which is has a
discrete state. States are represented by integers (0, 1, ... N).

In addition, every active link has an *orientation code* and a *link state
code*. The orientation code represents the orientation of the link in space: is
it "vertical" (aligned with the y axis), "horizontal" (aligned with x), or in
some other orientation? The number of possible orientations depends on the
subclass. The base class has only one orientation code (0) (meaning
"orientation doesn't matter), but this is overridden in some of the subclasses.
For example, the OrientedRasterLCA has two orientation codes (0 and 1, for
vertical and horizontal), while the OrientedHexLCA has three (representing the
three axes in a hex-cell / triagonal grid).

Each active link also has a *link state code*. The *state* of a link refers to
its particular combination of nodes and its orientation. For example, link
state 1 refers to a link in which the tail-node has state 0, the head-node has
state 1, and the orientation code is 0. The number of possible link states is
equal to R N^2, where R is the number of orientations (1 to 3, depending on the
subclass) and N is the number of possible node states. The simplest possible
Landlab CA model would have just one orientation code and two possible cell
states, so that there are four unique link states. These would be represented
by the tuples of (tail-node state, head-node state, orientation) as follows::

    link state 0 = (0, 0, 0)
    link state 1 = (0, 1, 0)
    link state 2 = (1, 0, 0)
    link state 3 = (1, 1, 0)

Main data structures
--------------------
node_state : 1d array (x number of nodes in grid)
    Node-based grid of node-state codes. This is the grid of cell (sic) states.

link_state_dict : dictionary
    Keys are 3-element tuples that represent the cell-state pairs and
    orientation code for each possible link type; values are the corresponding
    link-state codes. Allows you to look up the link-state code corresponding
    to a particular pair of adjacent nodes with a particular orientation.

node_pair : list (x number of possible link states)
    List of 3-element tuples representing all the various link states. Allows
    you to look up the node states and orientation corresponding to a
    particular link-state ID.

priority_queue : heap containing event records
    Queue containing all future transition events, sorted by time of occurrence
    (from soonest to latest).

next_update : 1d array (x number of active links)
    Time (in the future) at which the link will undergo its next transition.
    You might notice that the update time for every scheduled transition is
    also stored in each Event object in the event queue. Why store it twice?
    Because a scheduled event might be invalidated after the event has been
    scheduled (because another transition has changed one of a link's two
    nodes, for example). The way to tell whether a scheduled event is still
    valid is to compare its time with the corresponding transition time in the
    *next_update* array. If they are different, the event is discarded.

link_orientation : 1d array of ints (x number of active links)
    Orientation code for each link.

link_state : 1d array of ints (x number of active links)
    State code for each link.

n_xn : 1d array of ints (x number of possible link states)
    Number of transitions ("xn" stands for "transition") from a given link
    state.

xn_to : 2d array of ints (# possible link states x max. # transitions)
    Stores the link-state code(s) to which a particular link state can
    transition. "max. # transitions" means the maximum number of transitions
    from a single state. For example, if each link state is associated with one
    and only one transition, then the maximum is 1, but if there is at least
    one link state that can have either of two different transitions, then the
    maximum would be two.

xn_rate : 2d array of floats (# possible link states x max. # transitions)
    Rate associated with each link-state transition.


Created GT Sep 2014, starting from link_cap.py.
"""
from __future__ import print_function

#X from _heapq import heappush
#X from _heapq import heappop
#X from _heapq import heapify
import landlab
import numpy as np
import pylab as plt

#X _USE_CYTHON = False

_CYTEST = True

_RUN_NEW = True

_TESTING = True

if _TESTING:
    from .cfuncs import PriorityQueue

#Xif _USE_CYTHON:
#    from .cfuncs import (update_link_states_and_transitions,
#                         run_cts, run_cts_lean, PriorityQueue)

# TODO:
#   - REMOVE TEST CLAUSE (MAKE IT NORMAL BEHAVIOR)
#   - REMOVE OBSOLUTE, UNUSED METHODS
#   - RENAME "_new" METHODS (REMOVE THE "_new")
#   - UPDATE CALLING SYNTAX IN CELLLAB_CTS AND FRIENDS
if _CYTEST:
    from landlab.ca.cfuncs import (update_node_states,
                                   push_transitions_to_event_queue_new,
                                   do_transition_new,
                                   update_link_states_and_transitions_new,
                                   run_cts_new, get_next_event_new)

_NEVER = 1e50

_DEBUG = False

_TEST = False

_CORE = landlab.grid.base.CORE_NODE


class Transition(object):

    """A transition from one state to another.

    Represents a transition from one state ("from_state") to another
    ("to_state") at a link. The transition probability is represented by a rate
    parameter "rate", with dimensions of 1/T. The probability distribution of
    time until the transition event occurs is exponentional with mean 1/rate.
    The optional name parameter allows the caller to assign a name to any given
    transition.

    Note that from_state and to_state can now be either integer IDs for the
    standardised ordering of the link states (as before), or tuples explicitly
    describing the node state at each end, and the orientation.
    Orientation is 0: horizontal, L-R; 1: vertical, bottom-top.
    For such a tuple, order is (left/bottom, right/top, orientation).

    Transition() constructor sets 3 required properties and 2 optional
    properties for a transition from one cell pair to another.

    Parameters
    ----------
    from_state : int
        Code for the starting state of the cell pair (link)
    to_state : int
        Code for the new state of the cell pair (link)
    rate : float
        Average rate at which this transition occurs (dimension of 1/time)
    name : string (optional)
        Name for this transition
    swap_properties : bool (optional)
        Flag: should properties be exchanged between the two cells?
    """

    def __init__(self, from_state, to_state, rate, name=None,
                 swap_properties=False, prop_update_fn=None):
        """
        Transition() constructor sets 3 required properties and 2 optional
        properties for a transition from one cell pair to another.

        Parameters
        ----------
        from_state : int
            Code for the starting state of the cell pair (link)
        to_state : int
            Code for the new state of the cell pair (link)
        rate : float
            Average rate at which this transition occurs (dimension of 1/time)
        name : string (optional)
            Name for this transition
        swap_properties : bool (optional)
            Flag: should properties be exchanged between the two cells?
        """
        self.from_state = from_state
        self.to_state = to_state
        self.rate = rate
        self.name = name
        self.swap_properties = swap_properties
        self.prop_update_fn = prop_update_fn


#Xif _USE_CYTHON:
#    from .cfuncs import Event
#else:
#    class Event(object):
#        """Transition event at a link.
#
#        Represents a transition event at a link. The transition occurs at a
#        given link and a given time, and it involves a transition into the
#        state xn_to (an integer code representing the new link state;
#        "xn" is shorthand for "transition").
#    
#        The class overrides the __lt__ (less than operator) method so that when
#        Event() objects are placed in a PriorityQueue, the earliest event is
#        given the highest priority (i.e., placed at the top of the queue).
#    
#        Event() constructor sets 3 required properties and one optional
#        property.
#    
#        Parameters
#        ----------
#        time : float
#            Time at which the event is scheduled to occur
#        link : int
#            ID of the link at which event occurs
#        xn_to : int
#            New state to which this cell pair (link) will transition
#        propswap : bool (optional)
#            Flag: does this event involve an exchange of properties between
#            the two cells?
#    
#        Examples
#        --------
#        >>> from landlab.ca.celllab_cts import Event
#        >>> e1 = Event( 10.0, 1, 2)
#        >>> e2 = Event( 2.0, 3, 1)
#        >>> e1 < e2
#        False
#        >>> e2 < e1
#        True
#        """
#    
#        def __init__(self, time, link, xn_to, propswap=False, prop_update_fn=None):
#            """
#            Event() constructor sets 3 required properties and one optional
#            property.
#    
#            Parameters
#            ----------
#            time : float
#                Time at which the event is scheduled to occur
#            link : int
#                ID of the link at which event occurs
#            xn_to : int
#                New state to which this cell pair (link) will transition
#            propswap : bool (optional)
#                Flag: does this event involve an exchange of properties between
#                the two cells?
#            """
#            self.time = time
#            self.link = link
#            self.xn_to = xn_to
#            self.propswap = propswap
#            self.prop_update_fn = prop_update_fn
#
#        def __lt__(self, other):
#            """
#            Overridden less-than operator: returns true if the event on the left
#            has an earlier scheduled time than the event on the right
#            """
#            return self.time < other.time


class CAPlotter(object):

    """Handle display of a CellLab-CTS grid.

    CAPlotter() constructor keeps a reference to the CA model, and
    optionally a colormap to be used with plots.

    Parameters
    ----------
    ca : LandlabCellularAutomaton object
        Reference to a CA model
    cmap : Matplotlib colormap, optional
        Colormap to be used in plotting
    """

    def __init__(self, ca, cmap=None, **kwds):
        """
        CAPlotter() constructor keeps a reference to the CA model, and
        optionally a colormap to be used with plots.

        Parameters
        ----------
        ca : LandlabCellularAutomaton object
            Reference to a CA model
        cmap : Matplotlib colormap, optional
            Colormap to be used in plotting
        """
        import matplotlib

        # Set the colormap; default to matplotlib's "jet" colormap
        if cmap is None:
            self._cmap = matplotlib.cm.jet
        else:
            self._cmap = cmap

        # Keep a reference to the CA model
        self.ca = ca

        # Initialize the plot and remember the grid type
        plt.ion()
        plt.figure(1)
        if type(ca.grid) is landlab.grid.hex.HexModelGrid:
            self.gridtype = 'hex'
        else:
            self.gridtype = 'rast'

    def update_plot(self):
        """Plot the current node state grid."""
        plt.clf()
        if self.gridtype == 'rast':
            nsr = self.ca.grid.node_vector_to_raster(self.ca.node_state)
            plt.imshow(nsr, interpolation='None',
                       origin='lower', cmap=self._cmap)
        else:
            self.ca.grid.hexplot(self.ca.node_state, color_map=self._cmap)

        plt.draw()
        plt.pause(0.001)

    def finalize(self):
        """Wrap up plotting.

        Wrap up plotting by switching off interactive model and showing the
        plot.
        """
        plt.ioff()
        plt.show()


class CellLabCTSModel(object):

    """Link-type (or doublet-type) cellular automaton model.

    A CellLabCTSModel implements a link-type (or doublet-type) cellular
    automaton model. A link connects a pair of cells. Each cell has a state
    (represented by an integer code), and each link also has a state that is
    determined by the states of the cell pair.

    Parameters
    ----------
    model_grid : Landlab ModelGrid object
        Reference to the model's grid
    node_state_dict : dict
        Keys are node-state codes, values are the names associated with
        these codes
    transition_list : list of Transition objects
        List of all possible transitions in the model
    initial_node_states : array of ints (x number of nodes in grid)
        Starting values for node-state grid
    prop_data : array (x number of nodes in grid), optional
        Array of properties associated with each node/cell
    prop_reset_value : number or object, optional
        Default or initial value for a node/cell property (e.g., 0.0).
        Must be same type as *prop_data*.
    """

    def __init__(self, model_grid, node_state_dict, transition_list,
                 initial_node_states, prop_data=None, prop_reset_value=None,
                 seed=0):
        """Initialize the CA model.

        Parameters
        ----------
        model_grid : Landlab ModelGrid object
            Reference to the model's grid
        node_state_dict : dict
            Keys are node-state codes, values are the names associated with
            these codes
        transition_list : list of Transition objects
            List of all possible transitions in the model
        initial_node_states : array of ints (x number of nodes in grid)
            Starting values for node-state grid
        prop_data : array (x number of nodes in grid), optional
            Array of properties associated with each node/cell
        prop_reset_value : number or object, optional
            Default or initial value for a node/cell property (e.g., 0.0).
            Must be same type as *prop_data*.
        seed : int, optional
            Seed for random number generation.
        """
        # Are we calling this from a subclass __init__? If so, then the
        # variable self.number_of_orientations should already be defined.
        try:
            self.number_of_orientations == 1
        except AttributeError:
            # if self.number_of_orientations not already defined
            self.number_of_orientations = 1

        # Keep a copy of the model grid; remember how many active links in it
        self.grid = model_grid

        # Initialize random number generation
        np.random.seed(seed)

        # Create an array that knows which links are connected to a boundary
        # node
        self.bnd_lnk = np.zeros(self.grid.number_of_links, dtype=np.int8)
        for link_id in range(self.grid.number_of_links):
            if self.grid.status_at_node[self.grid.node_at_link_tail[link_id]] != _CORE or self.grid.status_at_node[self.grid.node_at_link_head[link_id]] != _CORE:
                self.bnd_lnk[link_id] = True

        # Set up the initial node-state grid
        self.set_node_state_grid(initial_node_states)

        # Current simulation time starts out at zero
        self.current_time = 0.0

        # Figure out how many states there are, and make sure the input data
        # are self consistent.
        #   There are 2 x (N^2) link states, where N is the number of node
        # states. For example, if there are just two node states, 0 and 1, then
        # the possible oriented link pairs are listed below:
        #   0-0 0-1 1-0 1-1  0 0 1 1
        #                    0 1 0 1
        self.num_node_states = len(node_state_dict)
        self.num_node_states_sq = self.num_node_states * self.num_node_states
        self.num_link_states = (self.number_of_orientations *
                                self.num_node_states_sq)

        assert type(transition_list) is list, 'transition_list must be a list!'
        assert (transition_list), \
            'Transition list must contain at least one transition'
        last_type = None
        for t in transition_list:
            try:
                assert (t.from_state < self.num_link_states), \
                    'Transition from_state out of range'
                assert (t.to_state < self.num_link_states), \
                    'Transition to_state out of range'
                this_type = int
            # TODO: make orientation optional for cases where
            # self.number_of_orientations = 1
            except:
                # added to allow from and to states to be tuples, not just ids
                assert type(t.from_state) == tuple, \
                        'Transition from_state out of range'
                assert type(t.to_state) == tuple, \
                        'Transition to_state out of range'
                for i in t.from_state[:-1]:
                    assert (i < self.num_node_states), \
                        'Transition from_state out of range'
                for i in t.to_state[:-1]:
                    assert (i < self.num_node_states), \
                        'Transition to_state out of range'
                assert t.from_state[-1] < self.number_of_orientations, \
                    'Encoding for orientation in from_state must be < number of orientations.'
                assert t.to_state[-1] < self.number_of_orientations, \
                    'Encoding for orientation in to_state must be < number of orientations.'
                this_type = tuple
            assert last_type == this_type or last_type == None, \
                'All transition types must be either int IDs, or all tuples.'
            # this test to ensure all entries are either IDs, or tuples, not
            # mixed
            last_type = this_type

        # Create priority queue for events and next_update array for links
        #X self.event_queue = []
        #X heapify(self.event_queue)
        self.next_update = self.grid.add_zeros('link', 'next_update_time')
        self.priority_queue = PriorityQueue()
        self.next_trn_id = -np.ones(self.grid.number_of_links, dtype=np.int)

        # Assign link types from node types
        self.create_link_state_dict_and_pair_list()

        # DEJH adds: convert transition_list to IDs if necessary
        # This is the new part that allows Transition from_ and to_ types
        # to be specified either as ints, or as tuples.
        transition_list_as_ID = transition_list[:]
        if type(transition_list[0].from_state) == tuple:
            #(then they all are..., because of the assertions in __init__)
            for i in range(len(transition_list)):
                transition_list_as_ID[i].from_state = self.link_state_dict[
                    transition_list[i].from_state]
                transition_list_as_ID[i].to_state = self.link_state_dict[
                    transition_list[i].to_state]

        # Set up the information needed to determine the orientation of links
        # in the lattice. The default method just creates an array of zeros
        # (all orientations considered the same), but this will be overridden
        # in subclasses that do use orientation.
        self.setup_array_of_orientation_codes()

        # Using the grid of node states, figure out all the link states
        self.assign_link_states_from_node_types()

        # Create transition data for links
        self.setup_transition_data(transition_list_as_ID)

        # Put the various transitions on the event queue
        if _CYTEST:
            push_transitions_to_event_queue_new(len(self.grid.active_links),
                                                self.grid.active_links,
                                                self.n_trn, self.link_state,
                                                self.trn_id, self.trn_rate,
                                                self.next_update,
                                                self.next_trn_id,
                                                self.priority_queue)
        elif _RUN_NEW:
            self.push_transitions_to_event_queue_new()
        else:
            self.push_transitions_to_event_queue()

        # In order to keep track of cell "properties", we create an array of
        # indices that refer to locations in the caller's code where properties
        # are tracked.
        self.propid = np.arange(self.grid.number_of_nodes)
        if prop_data is None:
            self.prop_data = np.zeros(self.grid.number_of_nodes)
            self.prop_reset_value = 0.0
        else:
            self.prop_data = prop_data
            self.prop_reset_value = prop_reset_value
            
        # Determine and remember whether we will handle property swaps and/or
        # callbacks in this model.
        if np.amax(self.xn_propswap) > 0:
            self._use_propswap_or_callback = True
        else:
            self._use_propswap_or_callback = False

    def set_node_state_grid(self, node_states):
        """Set the grid of node-state codes to node_states.

        Sets the grid of node-state codes to node_states. Also checks
        to make sure node_states is in the proper format, which is to
        say, it's a Numpy array of the same length as the number of nodes in
        the grid.

        **Creates**:

        *  self.node_state : 1D array of ints (x number of nodes in grid)
           The node-state array

        Parameters
        ----------
        node_states : 1D array of ints (x number of nodes in grid)

        Notes
        -----
        The node-state array is attached to the grid as a field with the name
        'node_state'.
        """
        assert (type(node_states) is np.ndarray), \
            'initial_node_states must be a Numpy array'
        assert (len(node_states) == self.grid.number_of_nodes), \
            'length of initial_node_states must equal number of nodes in grid'
        self.grid.at_node['node_state'] = node_states
        self.node_state = node_states

    def create_link_state_dict_and_pair_list(self):
        """Create a dict of link-state to node-state.

        Creates a dictionary that can be used as a lookup table to find out
        which link state corresponds to a particular pair of node states. The
        dictionary keys are 3-element tuples, each of which represents the
        state of the TAIL node, the HEAD node, and the orientation of the link.
        The values are integer codes representing the link state numbers.

        Notes
        -----
        Performance note: making self.node_pair a tuple does not appear to
        change time to lookup values in update_node_states. Changing it to a
        2D array of int actually slows it down.
        """
        self.link_state_dict = {}
        self.node_pair = []
        k = 0
        for orientation in range(self.number_of_orientations):
            for tail_state in range(self.num_node_states):
                for head_state in range(self.num_node_states):
                    self.link_state_dict[
                        (tail_state, head_state, orientation)] = k
                    self.node_pair.append(
                        (tail_state, head_state, orientation))
                    k += 1

        if False and _DEBUG:
            print()
            print('create_link_state_dict_and_pair_list(): dict is:')
            print((self.link_state_dict))
            print('  and the pair list is:')
            print((self.node_pair))

    def setup_array_of_orientation_codes(self):
        """Create array of active link orientation codes.

        Creates and configures an array that contain the orientation code for
        each active link (and corresponding cell pair).

        **creates**:

        * ``self.link_orientation`` : 1D numpy array

        Notes
        -----

        The setup varies depending on the type of LCA. The default is
        non-oriented, in which case we just have an array of zeros. Subclasses
        will override this method to handle lattices in which orientation
        matters (for example, vertical vs. horizontal in an OrientedRasterLCA).
        """
        self.link_orientation = np.zeros(
            self.grid.number_of_links, dtype=np.int8)

    def assign_link_states_from_node_types(self):
        """Assign link-state code for each link.

        Takes lists/arrays of "tail" and "head" node IDs for each link, and a
        dictionary that associates pairs of node states (represented as a
        3-element tuple, comprising the TAIL state, FROM state, and
        orientation) to link states.

        **creates**:

        * ``self.link_state`` : 1D numpy array
        """
        self.link_state = np.zeros(self.grid.number_of_links, dtype=int)

        for i in self.grid.active_links:
            orientation = self.link_orientation[i]
            node_pair = (self.node_state[self.grid.node_at_link_tail[i]],
                         self.node_state[self.grid.node_at_link_head[i]],
                         orientation)
            self.link_state[i] = self.link_state_dict[node_pair]

        if False and _DEBUG:
            print()
            print('assign_link_states_from_node_types(): the link state array is:')
            print((self.link_state))

    def setup_transition_data(self, xn_list):
        """Create transition data arrays."""

        # First, create an array that stores the number of possible transitions
        # out of each state.
        self.n_xn = np.zeros(self.num_link_states, dtype=int)
        for xn in xn_list:
            self.n_xn[xn.from_state] += 1
        self.n_trn = np.zeros(self.num_link_states, dtype=np.int)

        # Now, create arrays to hold the "to state" and transition rate for each
        # transition. These arrays are dimensioned N x M where N is the number
        # of states, and M is the maximum number of transitions from a single
        # state (for example if state 3 could transition either to state 1 or
        # state 4, and the other states only had one or zero possible
        # transitions, then the maximum would be 2).
        max_transitions = np.max(self.n_xn)
        self.trn_id = np.zeros(
            (self.num_link_states, max_transitions), dtype=np.int)
        num_transitions = len(xn_list)
        self.trn_to = np.zeros(num_transitions, dtype=np.int)
        self.trn_rate = np.zeros(num_transitions)
        self.trn_propswap = np.zeros(num_transitions, dtype=np.int8)
        self.trn_prop_update_fn = np.zeros(num_transitions, dtype=object)
        # OLD
        self.xn_to = np.zeros(
            (self.num_link_states, max_transitions), dtype=int)
        self.xn_rate = np.zeros((self.num_link_states, max_transitions))
        self.xn_propswap = np.zeros(
            (self.num_link_states, max_transitions), dtype=np.int8)
        self.xn_prop_update_fn = np.zeros(
            (self.num_link_states, max_transitions), dtype=object)

        for trn in range(num_transitions):
            self.trn_to[trn] = xn_list[trn].to_state
            self.trn_rate[trn] = xn_list[trn].rate
            self.trn_propswap[trn] = xn_list[trn].swap_properties
            if xn_list[trn].prop_update_fn is not None:
                self.trn_prop_update_fn[trn] = xn_list[trn].prop_update_fn
                self._use_propswap_or_callback = True
            from_state = xn_list[trn].from_state
            self.trn_id[from_state, self.n_trn[from_state]] = trn
            self.n_trn[from_state] += 1

        #OLD
        # Populate the "to" and "rate" arrays
        # reset this and then re-do (inefficient but should work)
        self.n_xn[:] = 0
        for xn in xn_list:
            from_state = xn.from_state
            self.xn_to[from_state][self.n_xn[from_state]] = xn.to_state
            self.xn_rate[from_state][self.n_xn[from_state]] = xn.rate
            self.xn_propswap[from_state][
                self.n_xn[from_state]] = xn.swap_properties
            if xn.prop_update_fn is not None:
                self.xn_prop_update_fn[from_state][
                    self.n_xn[from_state]] = xn.prop_update_fn
                self._use_propswap_or_callback = True
            self.n_xn[from_state] += 1

        if False and _DEBUG:
            print()
            print('setup_transition_data():')
            print(('  n_xn', self.n_xn))
            print(('  to:', self.xn_to))
            print(('  rate:', self.xn_rate))

    def current_link_state(self, link_id):
        """Get the current state of a link.

        Used to determine whether the link state at link *link_id* has changed
        due to an independent change in the node-state grid. Returns the
        current state of the link based on the states of its two end nodes;
        this can be compared to the entry in self.link_state to determine
        whether the state has changed.

        Parameters
        ----------
        link_id : int
            ID of the active link to test

        Returns
        -------
        int
            New link state code

        Notes
        -----
        Vectorizing this might yield some speed.
        """

        # Find out the states of the two nodes, and the orientation
        tail_node_state = self.node_state[self.grid.node_at_link_tail[link_id]]
        head_node_state = self.node_state[self.grid.node_at_link_head[link_id]]
        orientation = self.link_orientation[link_id]

        # Return the corresponding state code.
        return (orientation * self.num_node_states_sq +
                tail_node_state * self.num_node_states + head_node_state)

    def update_link_states_and_transitions(self, current_time):
        """
        Following an "external" change to the node state grid, updates link
        states where necessary and creates any needed events.

        Notes
        -----
        **Algorithm**::

            FOR each active link:
                if the actual node pair is different from the link's code:
                    change the link state to be correct
                    schedule an event
        """
        if _CYTEST:
            update_link_states_and_transitions_new(self.grid.active_links,
                                               self.node_state, 
                                               self.grid.node_at_link_tail,
                                               self.grid.node_at_link_head,
                                               self.link_orientation,
                                               self.bnd_lnk,
                                               self.link_state,
                                               self.n_trn,
                                               self.priority_queue,
                                               self.next_update,
                                               self.next_trn_id,
                                               self.trn_id, self.trn_rate,
                                               self.num_node_states,
                                               self.num_node_states_sq,
                                               current_time)       
#X       elif _USE_CYTHON:
#            update_link_states_and_transitions(self.grid.active_links,
#                                               self.node_state, 
#                                               self.grid.node_at_link_tail,
#                                               self.grid.node_at_link_head,
#                                               self.link_orientation,
#                                               self.bnd_lnk,
#                                               self.link_state,
#                                               self.n_xn,
#                                               self.event_queue,
#                                               self.next_update,
#                                               self.xn_to, self.xn_rate,
#                                               self.num_node_states,
#                                               self.num_node_states_sq,
#                                               current_time,
#                                               self.xn_propswap,
#                                               self.xn_prop_update_fn)
        else:
            for i in self.grid.active_links:
                current_state = self.current_link_state(i)
                if current_state != self.link_state[i]:
                    self.update_link_state(i, current_state, current_time)

    def update_link_states_and_transitions_new(self, current_time):
        """
        Following an "external" change to the node state grid, updates link
        states where necessary and creates any needed events.

        Notes
        -----
        **Algorithm**::

            FOR each active link:
                if the actual node pair is different from the link's code:
                    change the link state to be correct
                    schedule an event
        """
        if _CYTEST:
            update_link_states_and_transitions_new(self.grid.active_links,
                                               self.node_state, 
                                               self.grid.node_at_link_tail,
                                               self.grid.node_at_link_head,
                                               self.link_orientation,
                                               self.bnd_lnk,
                                               self.link_state,
                                               self.n_trn,
                                               self.priority_queue,
                                               self.next_update,
                                               self.next_trn_id,
                                               self.trn_id, self.trn_rate,
                                               self.num_node_states,
                                               self.num_node_states_sq,
                                               current_time)       
        else:
            for i in self.grid.active_links:
                current_state = self.current_link_state(i)
                if current_state != self.link_state[i]:
                    self.update_link_state_new(i, current_state, current_time)

#X    def get_next_event(self, link, current_state, current_time):
#        """Get the next event for a link.
#
#        Returns the next event for link with ID "link", which is in state
#        "current state".
#
#        Parameters
#        ----------
#        link : int
#            ID of the link
#        current_state : int
#            Current state code for the link
#        current_time : float
#            Current time in simulation (i.e., time of event just processed)
#
#        Returns
#        -------
#        Event object
#            The returned Event object contains the time, link ID, and type of
#            the next transition event at this link.
#
#        Notes
#        -----
#        If there is only one potential transition out of the current state, a
#        time for the transition is selected at random from an exponential
#        distribution with rate parameter appropriate for this transition.
#
#        If there are more than one potential transitions, a transition time is
#        chosen for each, and the smallest of these applied.
#
#        Assumes that there is at least one potential transition from the
#        current state.
#        """
#        assert (self.n_xn[current_state] > 0), \
#            'must have at least one potential transition'
#
#        # Find next event time for each potential transition
#        if self.n_xn[current_state] == 1:
#            xn_to = self.xn_to[current_state][0]
#            propswap = self.xn_propswap[current_state][0]
#            next_time = np.random.exponential(
#                1.0 / self.xn_rate[current_state][0])
#            if _DEBUG:
#                print('next_time for 1 xn is ' + str(next_time))
#            prop_update_fn = self.xn_prop_update_fn[current_state][0]
#        else:
#            next_time = _NEVER
#            xn_to = None
#            propswap = False
#            for i in range(self.n_xn[current_state]):
#                this_next = np.random.exponential(
#                    1.0 / self.xn_rate[current_state][i])
#                if _DEBUG:
#                    print('this_next for >1 xn is ' + str(this_next))
#                if this_next < next_time:
#                    next_time = this_next
#                    xn_to = self.xn_to[current_state][i]
#                    propswap = self.xn_propswap[current_state][i]
#                    prop_update_fn = self.xn_prop_update_fn[current_state][i]
#
#        # Create and setup event, and return it
#        my_event = Event(next_time + current_time, link,
#                         xn_to, propswap, prop_update_fn)
#
#        if _DEBUG:
#            print('get_next_event():')
#            print(('  current state:', current_state))
#            print(('  node pair:', self.node_pair[current_state]))
#            print(('  next_time:', my_event.time))
#            print(('  link:', my_event.link))
#            print(('  xn_to:', my_event.xn_to))
#            print(('  nxn:', self.n_xn[current_state]))
#            print(('  rate:', self.xn_rate[current_state][:]))
#            print(('  propswap:', my_event.propswap))
#
#        return my_event

    def get_next_event_new(self, link, current_state, current_time):
        """Get the next event for a link.

        Returns the next event for link with ID "link", which is in state
        "current state".

        Parameters
        ----------
        link : int
            ID of the link
        current_state : int
            Current state code for the link
        current_time : float
            Current time in simulation (i.e., time of event just processed)

        Returns
        -------
        Event object
            The returned Event object contains the time, link ID, and type of
            the next transition event at this link.

        Notes
        -----
        If there is only one potential transition out of the current state, a
        time for the transition is selected at random from an exponential
        distribution with rate parameter appropriate for this transition.

        If there are more than one potential transitions, a transition time is
        chosen for each, and the smallest of these applied.

        Assumes that there is at least one potential transition from the
        current state.
        """
        assert (self.n_xn[current_state] > 0), \
            'must have at least one potential transition'

        # Find next event time for each potential transition: new version
        if self.n_trn[current_state] == 1:
            trn_id = self.trn_id[current_state, 0]
            next_time = np.random.exponential(1.0 / self.trn_rate[trn_id])
            if _DEBUG:
                print('rand (' + str(self.trn_rate[trn_id]) + ': ' + str(next_time))
        else:
            next_time = _NEVER
            trn_id = -1
            for i in range(self.n_trn[current_state]):
                this_next = np.random.exponential(
                    1.0 / self.trn_rate[self.trn_id[current_state][i]])
                if _DEBUG:
                    print('rand2 (' + str(self.trn_rate[self.trn_id[current_state][i]]) + ': ' + str(this_next))
                if this_next < next_time:
                    if _DEBUG:
                        print(' using it')
                    next_time = this_next
                    trn_id = self.trn_id[current_state, i]

        if _DEBUG:
            print('get_next_event_new():')
            print(('  current state:', current_state))
            print(('  node pair:', self.node_pair[current_state]))
            print(('  next_time:', next_time))
            print(('  link:', link))
            print(('  trn_id:', trn_id))

        return (next_time + current_time, trn_id)

#X    def push_transitions_to_event_queue(self):
#        """
#        Initializes the event queue by creating transition events for each
#        cell pair that has one or more potential transitions and pushing these
#        onto the queue. Also records scheduled transition times in the
#        self.next_update array.
#        
#        Examples
#        --------
#        >>> from landlab import RasterModelGrid
#        >>> from landlab.ca.celllab_cts import Transition
#        >>> from landlab.ca.oriented_raster_cts import OrientedRasterCTS
#        >>> import numpy as np
#        >>> grid = RasterModelGrid((3, 5))
#        >>> nsd = {0 : 'zero', 1 : 'one'}
#        >>> trn_list = []
#        >>> trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
#        >>> trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
#        >>> trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
#        >>> trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
#        >>> ins = np.arange(15) % 2
#        >>> cts = OrientedRasterCTS(grid, nsd, trn_list, ins)
#        """
#        if False and _DEBUG:
#            print(('push_transitions_to_event_queue():',
#                   self.num_link_states, self.n_xn))
#
#        for i in self.grid.active_links:
#
#            if self.n_xn[self.link_state[i]] > 0:
#                event = self.get_next_event(i, self.link_state[i], 0.0)
#                heappush(self.event_queue, event)
#                self.next_update[i] = event.time
#                
#            else:
#                self.next_update[i] = _NEVER
#
#        if False and _DEBUG:
#            print('  push_transitions_to_event_queue(): events in queue are now:')
#            for e in self.event_queue:
#                print('    next_time:', e.time, 'link:',
#                      e.link, 'xn_to:', e.xn_to)

    def push_transitions_to_event_queue_new(self):
        """
        Initializes the event queue by creating transition events for each
        cell pair that has one or more potential transitions and pushing these
        onto the queue. Also records scheduled transition times in the
        self.next_update array.
        
        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.ca.celllab_cts import Transition
        >>> from landlab.ca.oriented_raster_cts import OrientedRasterCTS
        >>> import numpy as np
        >>> grid = RasterModelGrid((3, 5))
        >>> nsd = {0 : 'zero', 1 : 'one'}
        >>> trn_list = []
        >>> trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
        >>> trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
        >>> ins = np.arange(15) % 2
        >>> cts = OrientedRasterCTS(grid, nsd, trn_list, ins)
        >>> ev0 = cts.priority_queue._queue[0]
        >>> np.round(100 * ev0[0])
        12.0
        >>> ev0[2]  # this is the link ID
        16
        >>> ev6 = cts.priority_queue._queue[6]
        >>> np.round(100 * ev6[0])
        27.0
        >>> ev6[2]  # this is the link ID
        6
        >>> cts.next_trn_id[ev0[2]]  # ID of the transition to occur at this link
        3
        >>> cts.next_trn_id[cts.grid.active_links]
        array([-1,  2, -1,  1,  0,  1,  0,  2, -1,  3])
        """
        if False and _DEBUG:
            print(('push_transitions_to_event_queue_new():',
                   self.num_link_states, self.n_xn))

        for i in self.grid.active_links:

            if self.n_xn[self.link_state[i]] > 0:
                (ev_time, trn_id) = self.get_next_event_new(i, self.link_state[i], 0.0)
                self.priority_queue.push(i, ev_time)
                self.next_update[i] = ev_time
                self.next_trn_id[i] = trn_id

            else:
                self.next_update[i] = _NEVER

    #@profile
    def update_node_states(self, tail_node, head_node, new_link_state):
        """Update the states of the two nodes in the given link.

        Parameters
        ----------
        tail_node : int
            ID of the tail node of the link (cell pair) in question
        head_node : int
            ID of the head node of the link (cell pair) in question
        new_link_state : int
            Link state code for the new cell pair

        Returns
        -------
        (bool, bool)
            Flags indicating whether the tail node and head node, respectively,
            have changed state
        """
        # Remember the previous state of each node so we can detect whether the
        # state has changed
        old_tail_node_state = self.node_state[tail_node]
        old_head_node_state = self.node_state[head_node]

        # Change to the new states
        if self.grid.status_at_node[tail_node] == _CORE:
            self.node_state[tail_node] = self.node_pair[new_link_state][0]
        if self.grid.status_at_node[head_node] == _CORE:
            self.node_state[head_node] = self.node_pair[new_link_state][1]

        return self.node_state[tail_node] != old_tail_node_state, \
               self.node_state[head_node] != old_head_node_state

#X    def update_link_state(self, link, new_link_state, current_time):
#        """
#        Implements a link transition by updating the current state of the link
#        and (if appropriate) choosing the next transition event and pushing it
#        on to the event queue.
#
#        Parameters
#        ----------
#        link : int
#            ID of the link to update
#        new_link_state : int
#            Code for the new state
#        current_time : float
#            Current time in simulation
#        """
#        if _DEBUG:
#            print('update_link_state() link ' + str(link) + ' to state ' + str(new_link_state) + ' at time ' + str(current_time))
#
#        # If the link connects to a boundary, we might have a different state
#        # than the one we planned
#        if self.bnd_lnk[link]:
#            fns = self.node_state[self.grid.node_at_link_tail[link]]
#            tns = self.node_state[self.grid.node_at_link_head[link]]
#            orientation = self.link_orientation[link]
#            new_link_state = orientation * self.num_node_states_sq + \
#                fns * self.num_node_states + tns
#
#        self.link_state[link] = new_link_state
#        if self.n_xn[new_link_state] > 0:
#            event = self.get_next_event(link, new_link_state, current_time)
#            heappush(self.event_queue, event)
#            self.next_update[link] = event.time
#        else:
#            self.next_update[link] = _NEVER

    def update_link_state_new(self, link, new_link_state, current_time):
        """
        Implements a link transition by updating the current state of the link
        and (if appropriate) choosing the next transition event and pushing it
        on to the event queue.

        Parameters
        ----------
        link : int
            ID of the link to update
        new_link_state : int
            Code for the new state
        current_time : float
            Current time in simulation
        """
        if _DEBUG:
            print('update_link_state_new() link ' + str(link) + ' to state '
                  + str(new_link_state) + ' at time ' + str(current_time))

        # If the link connects to a boundary, we might have a different state
        # than the one we planned
        if self.bnd_lnk[link]:
            fns = self.node_state[self.grid.node_at_link_tail[link]]
            tns = self.node_state[self.grid.node_at_link_head[link]]
            orientation = self.link_orientation[link]
            new_link_state = int(orientation * self.num_node_states_sq
                                 + fns * self.num_node_states + tns)

        self.link_state[link] = new_link_state
        if self.n_trn[new_link_state] > 0:
            if _CYTEST:
                (event_time, trn_id) = get_next_event_new(link,
                                                new_link_state, 
                                                current_time,
                                                self.n_trn,
                                                self.trn_id,
                                                self.trn_rate)
            else:
                (event_time, trn_id) = self.get_next_event_new(link, 
                                        new_link_state, current_time)
            self.priority_queue.push(link, event_time)
            if _DEBUG:
                print('Pushed event at ' + str(link) + ' for time '
                      + str(event_time) + ' id ' + str(trn_id))
            self.next_update[link] = event_time
            self.next_trn_id[link] = trn_id
        else:
            self.next_update[link] = _NEVER
            self.next_trn_id[link] = -1

    def do_transition(self, event, current_time, plot_each_transition=False,
                      plotter=None):
        """Transition state.

        Implements a state transition.

        Parameters
        ----------
        event : Event object
            Event object containing the data for the current transition event
        current_time : float
            Current time in simulation
        plot_each_transition : bool (optional)
            True if caller wants to show a plot of the grid after this
            transition
        plotter : CAPlotter object
            Sent if caller wants a plot after this transition

        Notes
        -----
        First checks that the transition is still valid by comparing the
        link's next_update time with the corresponding update time in the
        event object.

        If the transition is valid, we:

        1. Update the states of the two nodes attached to the link
        2. Update the link's state, choose its next transition, and push
           it on the event queue.
        3. Update the states of the other links attached to the two nodes,
           choose their next transitions, and push them on the event queue.
        """
        if _DEBUG:
            print()
            print('py do_transition() for link',event.link,'time',event.time, ' cur time ', current_time)
            print(np.count_nonzero(self.node_state != 0))
            print(self.link_state[event.link])

        # We'll process the event if its update time matches the one we have
        # recorded for the link in question. If not, it means that the link has
        # changed state since the event was pushed onto the event queue, and
        # in that case we'll ignore it.
        if event.time == self.next_update[event.link]:

            if _DEBUG:
                print('  event time =', event.time)

            tail_node = self.grid.node_at_link_tail[event.link]
            head_node = self.grid.node_at_link_head[event.link]
            tail_changed, head_changed = self.update_node_states(
                tail_node, head_node, event.xn_to)
            self.update_link_state(event.link, event.xn_to, event.time)

            # Next, when the state of one of the link's nodes changes, we have
            # to update the states of the OTHER links attached to it. This
            # could happen to one or both nodes.
            if tail_changed:

                if _DEBUG:
                    print(' tail node has changed state, so updating its links')
                    print(' links at node ' + str(tail_node) + ' are:')
                    print(self.grid.links_at_node[tail_node, :])
                    print(self.grid.active_link_dirs_at_node[tail_node, :])

                for i in range(self.grid.links_at_node.shape[1]):

                    link = self.grid.links_at_node[tail_node, i]
                    dir_code = self.grid.active_link_dirs_at_node[tail_node, i]
                    if _DEBUG:
                        print('tail checking link', link)
                        print('  dir code ' + str(dir_code) + ' event link ' + str(event.link))
                    if dir_code != 0 and link != event.link:

                        this_link_fromnode = self.grid.node_at_link_tail[link]
                        this_link_tonode = self.grid.node_at_link_head[link]
                        orientation = self.link_orientation[link]
                        current_pair = (self.node_state[this_link_fromnode],
                                        self.node_state[this_link_tonode],
                                        orientation)
                        new_link_state = self.link_state_dict[current_pair]
                        new_link_state2 = (
                            orientation * self.num_node_states_sq +
                            self.node_state[this_link_fromnode] * self.num_node_states +
                            self.node_state[this_link_tonode])
                        assert new_link_state == new_link_state2, 'oops'
                        self.update_link_state(
                            link, new_link_state, event.time)

            if head_changed:

                if _DEBUG:
                    print(' head node has changed state, so updating its links')

                for i in range(self.grid.links_at_node.shape[1]):

                    link = self.grid.links_at_node[head_node, i]
                    dir_code = self.grid.active_link_dirs_at_node[head_node, i]
                    if _DEBUG:
                        print('head checking link', link)
                    if dir_code != 0 and link != event.link:
                        this_link_fromnode = self.grid.node_at_link_tail[link]
                        this_link_tonode = self.grid.node_at_link_head[link]
                        orientation = self.link_orientation[link]
                        current_pair = (self.node_state[this_link_fromnode],
                                        self.node_state[this_link_tonode],
                                        orientation)
                        new_link_state = self.link_state_dict[current_pair]
                        new_link_state2 = (
                            orientation * self.num_node_states_sq +
                            self.node_state[this_link_fromnode] * self.num_node_states +
                            self.node_state[this_link_tonode])
                        assert new_link_state == new_link_state2, 'oops'
                        self.update_link_state(
                            link, new_link_state, event.time)

            # If requested, display a plot of the grid
            if plot_each_transition and (plotter is not None):
                plotter.update_plot()

            # If this event involves an exchange of properties (i.e., the
            # event involves motion of an object that posses properties we
            # want to track), implement the swap.
            #   If the event requires a call to a user-defined callback
            # function, we handle that here too.
            if event.propswap:
                tmp = self.propid[tail_node]
                self.propid[tail_node] = self.propid[head_node]
                self.propid[head_node] = tmp
                if self.grid.status_at_node[tail_node] != _CORE:
                    self.prop_data[self.propid[tail_node]
                                   ] = self.prop_reset_value
                if self.grid.status_at_node[head_node] != _CORE:
                    self.prop_data[self.propid[head_node]
                                   ] = self.prop_reset_value
                if event.prop_update_fn is not None:
                    event.prop_update_fn(
                        self, tail_node, head_node, event.time)

        elif _DEBUG:
            print('  event time is', event.time, 'but update time is',
                  self.next_update[event.link], 'so event will be ignored')

    def do_transition_new(self, event_link, event_time, current_time, plot_each_transition=False,
                      plotter=None):
        """Transition state.

        Implements a state transition.

        Parameters
        ----------
        event : Event object
            Event object containing the data for the current transition event
        current_time : float
            Current time in simulation
        plot_each_transition : bool (optional)
            True if caller wants to show a plot of the grid after this
            transition
        plotter : CAPlotter object
            Sent if caller wants a plot after this transition

        Notes
        -----
        First checks that the transition is still valid by comparing the
        link's next_update time with the corresponding update time in the
        event object.

        If the transition is valid, we:

        1. Update the states of the two nodes attached to the link
        2. Update the link's state, choose its next transition, and push
           it on the event queue.
        3. Update the states of the other links attached to the two nodes,
           choose their next transitions, and push them on the event queue.
           
        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.ca.celllab_cts import Transition
        >>> from landlab.ca.oriented_raster_cts import OrientedRasterCTS
        >>> import numpy as np
        >>> grid = RasterModelGrid((3, 5))
        >>> nsd = {0 : 'zero', 1 : 'one'}
        >>> trn_list = []
        >>> trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
        >>> trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
        >>> ins = np.arange(15) % 2
        >>> cts = OrientedRasterCTS(grid, nsd, trn_list, ins)
        >>> (tm, idx, link) = cts.priority_queue.pop()
        >>> np.round(100 * tm)
        12.0
        >>> idx
        6
        >>> link
        16
        >>> cts.grid.node_at_link_tail[link]
        8
        >>> cts.grid.node_at_link_head[link]
        13
        >>> cts.next_trn_id[link]
        3
        >>> cts.trn_to[cts.next_trn_id[link]]
        7
        >>> cts.do_transition_new(link, tm, 0.0)
        >>> cts.node_state[8]
        1
        >>> cts.node_state[13]
        1
        >>> cts.link_state[16]
        7
        >>> cts.next_update[16] == _NEVER
        True
        >>> cts.next_trn_id[16]
        -1
        """
        if _DEBUG:
            print()
            print('py do_transition_new() for link', event_link,'time',
                  event_time, ' cur time ', current_time)

        # We'll process the event if its update time matches the one we have
        # recorded for the link in question. If not, it means that the link has
        # changed state since the event was pushed onto the event queue, and
        # in that case we'll ignore it.
        if event_time == self.next_update[event_link]:

            if _DEBUG:
                print('  event time =', event_time)

            tail_node = self.grid.node_at_link_tail[event_link]
            head_node = self.grid.node_at_link_head[event_link]
            trn_id = self.next_trn_id[event_link]
            trn_to = self.trn_to[trn_id]
            if _CYTEST:
                old_tail_node_state = self.node_state[tail_node]
                old_head_node_state = self.node_state[head_node]
                update_node_states(self.node_state,
                       self.grid.status_at_node,
                       tail_node, 
                       head_node,
                       trn_to,
                       self.num_node_states)
                tail_changed = (old_tail_node_state != self.node_state[tail_node])
                head_changed = (old_head_node_state != self.node_state[head_node])
            else:
                tail_changed, head_changed = self.update_node_states(
                                                tail_node, head_node, trn_to)
            self.update_link_state_new(event_link, trn_to, event_time)

            # Next, when the state of one of the link's nodes changes, we have
            # to update the states of the OTHER links attached to it. This
            # could happen to one or both nodes.
            if tail_changed:

                if _DEBUG:
                    print(' tail node has changed state, so updating its links')
                    print(' links at node ' + str(tail_node) + ' are:')
                    print(self.grid.links_at_node[tail_node, :])
                    print(self.grid.active_link_dirs_at_node[tail_node, :])

                for i in range(self.grid.links_at_node.shape[1]):

                    link = self.grid.links_at_node[tail_node, i]
                    dir_code = self.grid.active_link_dirs_at_node[tail_node, i]
                    if _DEBUG:
                        print('tail checking link', link)
                        print('  dir code ' + str(dir_code) + ' event link ' + str(event_link))
                    if dir_code != 0 and link != event_link:

                        this_link_fromnode = self.grid.node_at_link_tail[link]
                        this_link_tonode = self.grid.node_at_link_head[link]
                        orientation = self.link_orientation[link]
                        current_pair = (self.node_state[this_link_fromnode],
                                        self.node_state[this_link_tonode],
                                        orientation)
                        new_link_state = self.link_state_dict[current_pair]
                        new_link_state2 = (
                            orientation * self.num_node_states_sq +
                            self.node_state[this_link_fromnode] * self.num_node_states +
                            self.node_state[this_link_tonode])
                        assert new_link_state == new_link_state2, 'oops'
                        self.update_link_state_new(
                                            link, new_link_state, event_time)

            if head_changed:

                if _DEBUG:
                    print(' head node has changed state, so updating its links')

                for i in range(self.grid.links_at_node.shape[1]):

                    link = self.grid.links_at_node[head_node, i]
                    dir_code = self.grid.active_link_dirs_at_node[head_node, i]
                    if _DEBUG:
                        print('head checking link', link)
                    if dir_code != 0 and link != event_link:
                        this_link_fromnode = self.grid.node_at_link_tail[link]
                        this_link_tonode = self.grid.node_at_link_head[link]
                        orientation = self.link_orientation[link]
                        current_pair = (self.node_state[this_link_fromnode],
                                        self.node_state[this_link_tonode],
                                        orientation)
                        new_link_state = self.link_state_dict[current_pair]
                        new_link_state2 = (
                            orientation * self.num_node_states_sq +
                            self.node_state[this_link_fromnode] * self.num_node_states +
                            self.node_state[this_link_tonode])
                        assert new_link_state == new_link_state2, 'oops'
                        self.update_link_state_new(
                                            link, new_link_state, event_time)

            # If requested, display a plot of the grid
            if plot_each_transition and (plotter is not None):
                plotter.update_plot()

            # If this event involves an exchange of properties (i.e., the
            # event involves motion of an object that posses properties we
            # want to track), implement the swap.
            #   If the event requires a call to a user-defined callback
            # function, we handle that here too.
            if self.trn_propswap[trn_id]:
                tmp = self.propid[tail_node]
                self.propid[tail_node] = self.propid[head_node]
                self.propid[head_node] = tmp
                if self.grid.status_at_node[tail_node] != _CORE:
                    self.prop_data[self.propid[tail_node]
                                   ] = self.prop_reset_value
                if self.grid.status_at_node[head_node] != _CORE:
                    self.prop_data[self.propid[head_node]
                                   ] = self.prop_reset_value
                if self.trn_prop_update_fn[trn_id] is not None:
                    self.trn_prop_update_fn[trn_id](
                        self, tail_node, head_node, event_time)

            if False and _DEBUG:
                n = self.grid.number_of_nodes
                for r in range(self.grid.number_of_node_rows):
                    for c in range(self.grid.number_of_node_columns):
                        n -= 1
                        print('{0:.0f}'.format(self.node_state[n]), end=' ')
                    print()
                if self.propid is not None:
                    print()
                    n = self.grid.number_of_nodes
                    for r in range(self.grid.number_of_node_rows):
                        for c in range(self.grid.number_of_node_columns):
                            n -= 1
                            print('{0:2.0f}'.format(self.propid[n]), end=' ')
                        print()

        elif _DEBUG:
            print('  event time is', event_time, 'but update time is',
                  self.next_update[event_link], 'so event will be ignored')

    def update_component_data(self, new_node_state_array):
        """Update all component data.

        Call this method to update all data held by the component, if, for
        example, another component or boundary conditions modify the node
        statuses outside the component between run steps.

        This method updates all necessary properties, including both node and
        link states.

        *new_node_state_array* is the updated list of node states, which must
        still all be compatible with the state list originally supplied to
        this component.
        """
        self.set_node_state_grid(new_node_state_array)
        self.assign_link_states_from_node_types()
        self.push_transitions_to_event_queue_new()

    #@profile
    def run(self, run_to, node_state_grid=None,
            plot_each_transition=False, plotter=None):
        """Run the model forward for a specified period of time.

        Parameters
        ----------
        run_to : float
            Time to run to, starting from self.current_time
        node_state_grid : 1D array of ints (x number of nodes) (optional)
            Node states (if given, replaces model's current node state grid)
        plot_each_transition : bool (optional)
            Option to display the grid after each transition
        plotter : CAPlotter object (optional)
            Needed if caller wants to plot after every transition
           
        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.ca.celllab_cts import Transition
        >>> from landlab.ca.oriented_raster_cts import OrientedRasterCTS
        >>> import numpy as np
        >>> grid = RasterModelGrid((3, 5))
        >>> nsd = {0 : 'zero', 1 : 'one'}
        >>> trn_list = []
        >>> trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
        >>> trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
        >>> ins = np.arange(15) % 2
        >>> cts = OrientedRasterCTS(grid, nsd, trn_list, ins)
        """
        if node_state_grid is not None:
            self.set_node_state_grid(node_state_grid)

#X        if plot_each_transition or self._use_propswap_or_callback:
#            lean_run = False
#        else:
#            lean_run = True

#X        if _USE_CYTHON and not lean_run:
#            self.current_time = run_cts(run_to, self.current_time,
#               plot_each_transition,
#               plotter,
#               self.event_queue,
#               self.next_update,                  
#               self.grid.node_at_link_tail,                  
#               self.grid.node_at_link_head,                  
#               self.node_state,            
#               self.link_state,
#               self.grid.status_at_node,
#               self.link_orientation,
#               self.propid,
#               self.prop_data,
#               self.n_xn,
#               self.xn_to,
#               self.xn_rate, 
#               self.grid.links_at_node,
#               self.grid.active_link_dirs_at_node,
#               self.num_node_states,
#               self.num_node_states_sq,
#               self.prop_reset_value,
#               self.xn_propswap,
#               self.xn_prop_update_fn,
#               self.bnd_lnk,
#               self)
#        elif _USE_CYTHON and lean_run:
#            self.current_time = run_cts_lean(run_to, self.current_time,
#               self.event_queue,
#               self.next_update,                  
#               self.grid.node_at_link_tail,                  
#               self.grid.node_at_link_head,                  
#               self.node_state,            
#               self.link_state,
#               self.grid.status_at_node,
#               self.link_orientation,
#               self.n_xn,
#               self.xn_to,
#               self.xn_rate, 
#               self.grid.links_at_node,
#               self.grid.active_link_dirs_at_node,
#               self.num_node_states,
#               self.num_node_states_sq,
#               self.bnd_lnk)

#X        if _RUN_NEW:
            
        self.current_time = run_cts_new(run_to, self.current_time,
                        self.priority_queue,
                        self.next_update,                  
                        self.grid.node_at_link_tail,                  
                        self.grid.node_at_link_head,                  
                        self.node_state,            
                        self.next_trn_id,            
                        self.trn_to,
                        self.grid.status_at_node,
                        self.num_node_states,
                        self.num_node_states_sq,
                        self.bnd_lnk,
                        self.link_orientation,
                        self.link_state,
                        self.n_trn,
                        self.trn_id,
                        self.trn_rate, 
                        self.grid.links_at_node,
                        self.grid.active_link_dirs_at_node,
                        self.trn_propswap,
                        self.propid,
                        self.prop_data,
                        self.prop_reset_value,
                        self.trn_prop_update_fn,
                        self,
                        plot_each_transition,
                        plotter)         
#X        else:
#
#            # Continue until we've run out of either time or events
#            while self.current_time < run_to and self.event_queue:
#    
#                if _DEBUG:
#                    print('Current Time = ', self.current_time)
#    
#                # Is there an event scheduled to occur within this run?
#                if self.event_queue[0].time <= run_to:
#    
#                    # If so, pick the next transition event from the event queue
#                    ev = heappop(self.event_queue)
#    
#                    if _DEBUG:
#                        print('Event:', ev.time, ev.link, ev.xn_to)
#        
#                    self.do_transition(ev, self.current_time,
#                                       plot_each_transition, plotter)
#
#                    # Update current time
#                    self.current_time = ev.time
#                
#                # If there is no event scheduled for this span of time, simply
#                # advance current_time to the end of the current run period.
#                else:
#                    self.current_time = run_to
#
#                if _DEBUG:
#                    print(self.node_state)
                    
    
    def run_new(self, run_to, plot_each_transition=False, plotter=None):
        """Test of new approach using priority queue."""
        
        # Continue until we've run out of either time or events
        while self.current_time < run_to and self.priority_queue._queue:

            if _DEBUG:
                print('Current Time = ', self.current_time)

            # Is there an event scheduled to occur within this run?
            if self.priority_queue._queue[0][0] <= run_to:

                # If so, pick the next transition event from the event queue
                (ev_time, ev_idx, ev_link) = self.priority_queue.pop()

                if _DEBUG:
                    print('Event:', ev_time, ev_link, self.trn_to[self.next_trn_id[ev_link]])
                
                if _CYTEST:
                    do_transition_new(ev_link, ev_time,
                                      self.priority_queue, self.next_update,                  
                                      self.grid.node_at_link_tail,                  
                                      self.grid.node_at_link_head,
                                      self.node_state,
                                      self.next_trn_id,
                                      self.trn_to,
                                      self.grid.status_at_node,
                                      self.num_node_states,
                                      self.num_node_states_sq,
                                      self.bnd_lnk,
                                      self.link_orientation,
                                      self.link_state,
                                      self.n_trn,
                                      self.trn_id,
                                      self.trn_rate, 
                                      self.grid.links_at_node,
                                      self.grid.active_link_dirs_at_node,
                                      self.trn_propswap,
                                      self.propid,
                                      self.prop_data,
                                      self.prop_reset_value,
                                      self.trn_prop_update_fn,
                                      self,
                                      plot_each_transition=False,
                                      plotter=None)
                else:
                    self.do_transition_new(ev_link, ev_time, self.current_time,
                                           plot_each_transition, plotter)

                # Update current time
                self.current_time = ev_time

            # If there is no event scheduled for this span of time, simply
            # advance current_time to the end of the current run period.
            else:
                self.current_time = run_to
                    
            if _DEBUG:
                print(self.node_state)

    def transition_info_as_string(self, event):
        """Returns info about a particular event as a string, for debug."""
        link = event[2]
        tail = self.grid.node_at_link_tail[link]
        head = self.grid.node_at_link_head[link]
        new_link_state = self.trn_to[self.next_trn_id[link]]
        new_tail_state = ((new_link_state / self.num_node_states)
                          % self.num_node_states) 
        new_head_state = new_link_state % self.num_node_states
        info_str = (str(event[0]) + ' '                       # sched time
                    + str(self.next_update[link]) + ' '       # sched time
                    + str(link) + ' '                         # link ID
                    + str(tail) + '=>'                        # tail ID
                    + str(head) + ' '                         # head ID
                    + str(self.link_orientation[link]) + ' '  # orientation
                    + str(self.node_state[tail]) + '=>'       # tail state
                    + str(self.node_state[head]) + ' '        # head state
                    + str(self.link_state[link]) + ' '        # link state
                    + str(self.next_trn_id[link]) + ' '       # trn ID
                    + str(new_link_state) + ' '               # new link state
                    + str(new_tail_state) + '=>'               # new tail state
                    + str(new_head_state)                     # new head state
                    )
        return info_str
                    

    def print_scheduled_transitions(self):
        """Display list of transitions in PQ, and related data, for debug."""
        print('tme tml lnk tln hdn orn tst hst tid nls nts nhs')
        for trn_event in self.priority_queue._queue:
            print(self.transition_info_as_string(trn_event))


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#! /usr/bin/env python
"""
Decorators for TheLandlab package.
"""

import inspect
import types

import six


def camel_case(text, sep=None):
    """
    Convert *text* to camel case. Use the *sep* keyword to specify the word
    separator. The default is to split on whitespace.

    >>> from landlab.framework.decorators import camel_case
    >>> camel_case('camel case')
    'CamelCase'
    >>> camel_case('camel_case', sep='_')
    'CamelCase'
    """
    return ''.join(text.title().split(sep))


class Error(Exception):
    """
    Exceptions for this module.
    """
    pass


class InterfaceImplementationError(Error):
    """
    Raise this error if the class does not implement an interface.
    """

    def __init__(self, cls, interface):
        self.cls = cls.__name__
        self.interface = interface.__name__

    def __str__(self):
        return "Class '%s' does not implement interface '%s'" % (
            self.cls, self.interface)


def is_implementation(cls, interface):
    """
    Check if *cls* implements *interface*. A class implements the interface
    class if it has the same members, the members have the same type, and
    methods have the same signature.

    Returns ``True`` if *cls* implements *interface*, otherwise ``False``.
    """
    for (name, value) in inspect.getmembers(interface):
        if isinstance(value, types.MethodType):
            try:
                cls_args = inspect.getargspec(getattr(cls, name))
                interface_args = inspect.getargspec(value)
            except AttributeError:
                six.print_('Missing attribute %s' % name)
                return False
            try:
                assert(len(cls_args.args) == len(interface_args.args))
            except AssertionError:
                six.print_('Mismatch in number of args for %s' % name)
                return False
        else:
            try:
                assert(type(getattr(cls, name)) ==
                       type(getattr(interface, name)))
            except (AttributeError, AssertionError):
                six.print_('Missing member or type mismatch for %s' % name)
                return False
    return True


class ImplementsOrRaise(object):
    """
    Decorator to indicate if a class implements interfaces. If the class
    does not implement the interface, raise an InterfaceImplementationError.
    If the class does implement the interface, decorate it with a
    __implements__ data mamber that is a tuple of the interfaces it implements.
    """

    def __init__(self, *interfaces):
        self._interfaces = interfaces

    def __call__(self, cls):
        cls.__implements__ = ()
        for interface in self._interfaces:
            if is_implementation(cls, interface):
                cls.__implements__ += (interface.__name__, )
            else:
                raise InterfaceImplementationError(cls, interface)
        return cls


class Implements(object):
    """
    Decorator to indicate if a class implements interfaces. Similar to the
    ImplementsOrRaise decorator except that this decorator silently ignores
    implemention errors. If the class does implement the interface, decorate
    it with a __implements__ data mamber that is a tuple of the interfaces it
    implements. Otherwise, don't do anything.
    """

    def __init__(self, *interfaces):
        self._interfaces = interfaces

    def __call__(self, cls):
        cls.__implements__ = ()
        for interface in self._interfaces:
            if is_implementation(cls, interface):
                cls.__implements__ += (interface.__name__, )
            else:
                pass
        return cls



import pytest
import numpy as np

from numpy.testing import assert_array_almost_equal

from landlab import RasterModelGrid
from landlab.components import StreamPowerSmoothThresholdEroder as Spst
from landlab.components import FlowAccumulator

def test_bad_nsp():

    mg = RasterModelGrid(4, 4, 1)
    mg.set_closed_boundaries_at_grid_edges(False, False, True, True)
    with pytest.raises(ValueError):
        spst = Spst(mg, K_sp = 1.0, n_sp = 1.01)

def test_no_thresh():
    K = 0.001
    U = 0.01
    m = 0.5
    n = 1.0
    threshold = 0.0
    dt = 1000

    mg = RasterModelGrid(30, 3, 100.)
    mg.set_closed_boundaries_at_grid_edges(True, False, True, False)
    z = mg.zeros(at='node')
    mg['node']['topographic__elevation'] = z + np.random.rand(len(z)) / 1000.

    fa = FlowAccumulator(mg)
    sp = Spst(mg, K_sp = K, threshold_sp=threshold)
    for i in range(100):
        fa.run_one_step()
        sp.run_one_step(dt)
        mg['node']['topographic__elevation'][mg.core_nodes] += U*dt

    actual_slopes = mg.at_node["topographic__steepest_slope"][mg.core_nodes[1:-1]]
    actual_areas = mg.at_node["drainage_area"][mg.core_nodes[1:-1]]

    predicted_slopes = (U / (K * (actual_areas ** m))) ** (1. / n)

    assert_array_almost_equal(actual_slopes, predicted_slopes)


def test_with_thresh():
    K = 0.001
    U = 0.01
    m = 0.5
    n = 1.0
    threshold = 1.0
    dt = 1000

    mg = RasterModelGrid(30, 3, 100.)
    mg.set_closed_boundaries_at_grid_edges(True, False, True, False)
    z = mg.zeros(at='node')
    mg['node']['topographic__elevation'] = z + np.random.rand(len(z)) / 1000.

    fa = FlowAccumulator(mg)
    sp = Spst(mg, K_sp = K, threshold_sp=threshold)
    for i in range(100):
        fa.run_one_step()
        sp.run_one_step(dt)
        mg['node']['topographic__elevation'][mg.core_nodes] += U*dt

    actual_slopes = mg.at_node["topographic__steepest_slope"][mg.core_nodes[1:-1]]
    actual_areas = mg.at_node["drainage_area"][mg.core_nodes[1:-1]]

    predicted_slopes_upper = ((U + threshold) / (K * (actual_areas ** m))) ** (1. / n)
    predicted_slopes_lower = ((U + 0.0) / (K * (actual_areas ** m))) ** (1. / n)

    # assert actual and predicted slopes are in the correct range for the slopes.
    assert np.all(actual_slopes > predicted_slopes_lower) == True
    assert np.all(actual_slopes < predicted_slopes_upper) == True



import numpy as np


class NodeGrid(object):
    """__init__((coord0, coord1))
    Create a grid of nodes.

    Parameters
    ----------
    coord0, coord1 : sequence of array-like
        Coordinates of grid nodes

    Returns
    -------
    NodeGrid :
        A newly-created NodeGrid

    Examples
    --------
    >>> from landlab.grid.unstructured.nodes import NodeGrid
    >>> ngrid = NodeGrid(([0, 0, 1, 1], [0, 1, 0, 1]))
    >>> ngrid.ndim == 2
    True
    >>> ngrid.number_of_nodes == 4
    True
    >>> ngrid.x
    array([ 0.,  1.,  0.,  1.])
    >>> ngrid.y
    array([ 0.,  0.,  1.,  1.])

    Create a 1D grid.

    >>> ngrid = NodeGrid(((0, 1, 3), ))
    >>> ngrid.ndim == 1
    True
    >>> ngrid.number_of_nodes
    3
    >>> ngrid.x
    array([ 0.,  1.,  3.])
    >>> ngrid.y
    Traceback (most recent call last):
    AttributeError: Grid has no y-coordinate
    """

    def __init__(self, nodes):
        """__init__((coord0, coord1))
        Create a grid of nodes.

        Parameters
        ----------
        coord0, coord1 : sequence of array-like
            Coordinates of grid nodes

        Returns
        -------
        NodeGrid :
            A newly-created NodeGrid

        Examples
        --------
        >>> from landlab.grid.unstructured.nodes import NodeGrid
        >>> ngrid = NodeGrid(([0, 0, 1, 1], [0, 1, 0, 1]))
        >>> ngrid.ndim == 2
        True
        >>> ngrid.number_of_nodes == 4
        True
        >>> ngrid.x
        array([ 0.,  1.,  0.,  1.])
        >>> ngrid.y
        array([ 0.,  0.,  1.,  1.])

        Create a 1D grid.

        >>> ngrid = NodeGrid(((0, 1, 3), ))
        >>> ngrid.ndim == 1
        True
        >>> ngrid.number_of_nodes
        3
        >>> ngrid.x
        array([ 0.,  1.,  3.])
        >>> ngrid.y
        Traceback (most recent call last):
        AttributeError: Grid has no y-coordinate
        """
        self._coords = np.vstack(
            [np.array(coord, dtype=float) for coord in nodes]
        )
        self._coords.flags['WRITEABLE'] = False

        self._number_of_nodes = len(nodes[0])

    @property
    def ndim(self):
        return self._coords.shape[0]

    @property
    def number_of_nodes(self):
        """
        Examples
        --------
        >>> from landlab.grid.unstructured.nodes import NodeGrid
        >>> ngrid = NodeGrid(([0, 0, 1], [0, 1, 0]))
        >>> ngrid.number_of_nodes
        3
        """
        return self._number_of_nodes

    @property
    def x(self):
        """Node coordinates of the "fast" dimension.

        Examples
        --------
        >>> from landlab.grid.unstructured.nodes import NodeGrid
        >>> ngrid = NodeGrid(([0, 0, 1], [0, 1, 0]))
        >>> ngrid.x
        array([ 0.,  1.,  0.])
        """
        return self._coords[-1]

    @property
    def y(self):
        """Node y-coordinates.

        Examples
        --------
        >>> from landlab.grid.unstructured.nodes import NodeGrid
        >>> ngrid = NodeGrid(([0, 0, 1], [0, 1, 0]))
        >>> ngrid.y
        array([ 0.,  0.,  1.])
        """
        try:
            return self._coords[-2]
        except IndexError:
            raise AttributeError('Grid has no y-coordinate')

    @property
    def z(self):
        """Node z-coordinates.

        Examples
        --------
        >>> from landlab.grid.unstructured.nodes import NodeGrid
        >>> ngrid = NodeGrid(([0, 0, 1], [0, 1, 0]))
        >>> ngrid.y
        array([ 0.,  0.,  1.])
        """
        try:
            return self._coords[-2]
        except IndexError:
            raise AttributeError('Grid has no z-coordinate')

    @property
    def coord(self):
        """
        Examples
        --------
        >>> from landlab.grid.unstructured.nodes import NodeGrid
        >>> ngrid = NodeGrid(([0, 0, 1], [0, 1, 0]))
        >>> ngrid.coord[0]
        array([ 0.,  0.,  1.])
        >>> ngrid.coord[1]
        array([ 0.,  1.,  0.])
        """
        return self._coords

    @property
    def point(self):
        """
        Examples
        --------
        >>> from landlab.grid.unstructured.nodes import NodeGrid
        >>> ngrid = NodeGrid(([0, 0, 1], [0, 1, 0]))
        >>> ngrid.point
        array([[ 0.,  0.],
               [ 0.,  1.],
               [ 1.,  0.]])
        >>> ngrid.point[1]
        array([ 0.,  1.])
        """
        return self._coords.T



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_number_of_corners():
    """Test number of corners on a raster."""
    grid = RasterModelGrid((4, 5))
    assert grid.number_of_corners == 12


def test_add_ones_at_corners():
    """Test add a field to corners."""
    grid = RasterModelGrid((4, 5))
    grid.add_ones('z', at='corner')

    assert_array_equal(grid.at_corner['z'],
                       [1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.])


def test_add_field_at_corners():
    """Test add a field to corners."""
    grid = RasterModelGrid((4, 5))
    x = np.arange(grid.number_of_corners)
    grid.add_field('z', x, at='corner')

    assert_array_equal(grid.at_corner['z'], x)



import pytest
import numpy as np

from landlab import RasterModelGrid
from landlab import BAD_INDEX_VALUE as XX
from landlab.components.sink_fill import SinkFiller


@pytest.fixture
def sink_grid1():
    """Create a 7x7 test grid with a well defined hole in it."""
    sink_grid = RasterModelGrid((7, 7), spacing=1.)

    z = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 2.0, 2.0, 2.0, 2.0, 2.0, 0.0],
            [0.0, 2.0, 1.6, 1.5, 1.6, 2.0, 0.0],
            [0.0, 2.0, 1.7, 1.6, 1.7, 2.0, 0.0],
            [0.0, 2.0, 1.8, 2.0, 2.0, 2.0, 0.0],
            [0.0, 1.0, 0.6, 1.0, 1.0, 1.0, 0.0],
            [0.0, 0.0, -0.5, 0.0, 0.0, 0.0, 0.0],
        ]
    ).flatten()

    sink_grid.add_field("topographic__elevation", z, at="node", units="-")

    sink_grid.outlet = 30
    sink_grid.lake_code = 17
    sink_grid.lake = np.array([16, 17, 18, 23, 24, 25])

    return sink_grid


@pytest.fixture
def sink_grid2():
    """
    Create a 10x10 test grid with a well defined hole in it, from a flat
    surface.
    """
    sink_grid = RasterModelGrid((10, 10), spacing=1.)

    lake = np.array([44, 45, 46, 54, 55, 56, 64, 65, 66])

    z = np.ones(100, dtype=float)
    z[lake] = 0.

    sink_grid.add_field("topographic__elevation", z, at="node", units="-")
    sink_grid.lake = lake

    return sink_grid


@pytest.fixture
def sink_grid3():
    """
    Create a 10x10 test grid with two well defined holes in it, into an
    inclined surface.
    """
    sink_grid = RasterModelGrid((10, 10), spacing=1.)

    lake1 = np.array([34, 35, 36, 44, 45, 46, 54, 55, 56])
    lake2 = np.array([77, 78, 87, 88])
    guard_nodes = np.array([23, 33, 53, 63])
    lake = np.concatenate((lake1, lake2))

    z = np.ones(100, dtype=float)
    # add slope
    z += sink_grid.node_x
    z[guard_nodes] += 0.001
    z[lake] = 0.

    sink_grid.add_field("node", "topographic__elevation", z, units="-")
    sink_grid.lake1 = lake1
    sink_grid.lake2 = lake2

    return sink_grid


@pytest.fixture
def sink_grid4():
    """
    Create a 10x10 test grid with two well defined holes in it, into an
    inclined surface. This time, one of the holes is a stupid shape, which
    will require the component to arrange flow back "uphill".
    """
    sink_grid = RasterModelGrid((10, 10), spacing=1.)

    lake1 = np.array([34, 35, 36, 44, 45, 46, 54, 55, 56, 65, 74])
    lake2 = np.array([78, 87, 88])
    guard_nodes = np.array([23, 33, 53, 63, 73, 83])
    lake = np.concatenate((lake1, lake2))
    # outlet = 35  # shouldn't be needed
    # outlet_array = np.array([outlet])

    z = np.ones(100, dtype=float)
    # add slope
    z += sink_grid.node_x
    z[guard_nodes] += 0.001  # forces the flow out of a particular node
    z[lake] = 0.

    # depr_outlet_target = np.empty(100, dtype=float)
    # depr_outlet_target.fill(XX)
    # depr_outlet_target = XX  # not well defined in this simplest case...?

    sink_grid.add_field("node", "topographic__elevation", z, units="-")
    sink_grid.lake1 = lake1
    sink_grid.lake2 = lake2

    # fr = FlowRouter(mg)

    return sink_grid


@pytest.fixture
def sink_grid5():
    """
    Create a 10x10 test grid with two well defined holes in it, into an
    inclined surface. This time, one of the holes is a stupid shape, which
    will require the component to arrange flow back "uphill". Exactly as
    V4, but this version tests D4 routing.

    Notes
    -----
    Here is the elevation grid::

    1.      2.      3.      4.      5.      6.      7.      8.      9.     10.
    1.      2.      3.      4.      5.      6.      7.      8.      9.     10.
    1.      2.      3.      4.001   5.      6.      7.      8.      9.     10.
    1.      2.      3.      4.001   0.      0.      0.      8.      9.     10.
    1.      2.      3.      4.      5.      6.      7.      8.      9.     10.
    1.      2.      3.      4.001   0.      0.      0.      8.      9.     10.
    1.      2.      3.      4.001   5.      0.      7.      8.      9.     10.
    1.      2.      3.      4.001   0.      6.      7.      8.      0.     10.
    1.      2.      3.      4.001   5.      6.      7.      0.      0.     10.
    1.      2.      3.      4.      5.      6.      7.      8.      9.     10.
    """
    sink_grid = RasterModelGrid((10, 10), spacing=1.)

    lake1 = np.array([34, 35, 36, 44, 45, 46, 54, 55, 56, 65, 74])
    lake2 = np.array([78, 87, 88])
    guard_nodes = np.array([23, 33, 53, 63, 73, 83])
    lake = np.concatenate((lake1, lake2))
    # outlet = 35  # shouldn't be needed
    # outlet_array = np.array([outlet])

    z = np.ones(100, dtype=float)
    # add slope
    z += sink_grid.node_x
    z[guard_nodes] += 0.001  # forces the flow out of a particular node
    z[lake] = 0.

    # depr_outlet_target = np.empty(100, dtype=float)
    # depr_outlet_target.fill(XX)
    # depr_outlet_target = XX  # not well defined in this simplest case...?

    sink_grid.add_field("node", "topographic__elevation", z, units="-")
    sink_grid.lake1 = lake1
    sink_grid.lake2 = lake2

    # fr = FlowRouter(mg)
    return sink_grid



import pytest
import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_add_field_at_node():
    """Add field at nodes."""
    grid = RasterModelGrid((4, 5))
    grid.add_field('z', np.arange(20), at='node')

    assert_array_equal(grid.at_node['z'], np.arange(20))


def test_add_field_without_at_keyword():
    """Test default is at nodes."""
    grid = RasterModelGrid((4, 5))
    grid.add_field('z', np.arange(20))

    assert_array_equal(grid.at_node['z'], np.arange(20))


def test_add_field_without_at():
    """Test raises error with wrong size array."""
    grid = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        grid.add_field('z', np.arange(21))


def test_add_field_at_grid():
    """Test add field at grid."""
    grid = RasterModelGrid((4, 5))
    grid.at_grid['value']=1
    assert_array_equal(1, grid.at_grid['value'].size)


def test_adding_field_at_grid_two_ways():
    """Test add field at grid two ways."""
    grid = RasterModelGrid((4, 5))
    grid.at_grid['value_1']=1
    grid.add_field('value_2', 1, at='grid')
    assert_array_equal(grid.at_grid['value_1'], grid.at_grid['value_2'])


def test_add_ones_zeros_empty_to_at_grid():
    """Test different add methods for keyword at='grid'"""
    grid = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        grid.add_zeros('value', at='grid')
    with pytest.raises(ValueError):
        grid.add_empty('value', at='grid')
    with pytest.raises(ValueError):
        grid.add_ones('value', at='grid')


def test_ones_zeros_empty_to_at_grid():
    """Test get array with field size methods for keyword at='grid'"""
    grid = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        grid.zeros(at='grid')
    with pytest.raises(ValueError):
        grid.empty(at='grid')
    with pytest.raises(ValueError):
        grid.ones(at='grid')



#! /usr/bin/env python
import numpy as np

from .nodestatus import (CLOSED_BOUNDARY, CORE_NODE, FIXED_GRADIENT_BOUNDARY,
                         FIXED_VALUE_BOUNDARY)
from ..utils.decorators import (cache_result_in_object,
                                make_return_array_immutable)

# Define the link types

#: Indicates a link is *active*, and can carry flux
ACTIVE_LINK = 0

#: Indicates a link has a fixed (gradient) value, & behaves as a boundary
FIXED_LINK = 2

#: Indicates a link is *inactive*, and cannot carry flux
INACTIVE_LINK = 4

LINK_STATUS_FLAGS_LIST = [
    ACTIVE_LINK,
    FIXED_LINK,
    INACTIVE_LINK,
]
LINK_STATUS_FLAGS = set(LINK_STATUS_FLAGS_LIST)


def is_fixed_link(node_status_at_link):
    """Find links that are fixed.

    A link is fixed if it connects a core node with a fixed value
    boundary node.

    Parameters
    ----------
    node_status_at_link : ndarray of int, shape `(n_links, 2)`
        Node status a link tail and head.

    Returns
    -------
    ndarray of bool, shape `(n_links, )`
        True if link is fixed.

    Examples
    --------
    >>> from landlab.grid.diagonals import is_fixed_link
    >>> from landlab import CORE_NODE, FIXED_GRADIENT_BOUNDARY
    >>> is_fixed_link([CORE_NODE, FIXED_GRADIENT_BOUNDARY])
    array([ True], dtype=bool)

    >>> from landlab import FIXED_VALUE_BOUNDARY
    >>> is_fixed_link([CORE_NODE, FIXED_VALUE_BOUNDARY])
    array([False], dtype=bool)

    >>> is_fixed_link([[FIXED_GRADIENT_BOUNDARY, CORE_NODE],
    ...                [CORE_NODE, CORE_NODE]])
    array([ True, False], dtype=bool)
    """
    node_status_at_link = np.asarray(node_status_at_link).reshape((-1, 2))

    is_core_node = node_status_at_link == CORE_NODE
    is_fixed_gradient_node = node_status_at_link == FIXED_GRADIENT_BOUNDARY

    return ((is_core_node[:, 0] & is_fixed_gradient_node[:, 1]) |
            (is_fixed_gradient_node[:, 0] & is_core_node[:, 1]))


def is_inactive_link(node_status_at_link):
    """Find links that are inactive.

    A link is inactive if it connects two boundary nodes or one of
    its nodes is closed.

    Parameters
    ----------
    node_status_at_link : ndarray of int, shape `(n_links, 2)`
        Node status a link tail and head.

    Returns
    -------
    ndarray of bool, shape `(n_links, )`
        True if link is isactive.

    Examples
    --------
    >>> from landlab.grid.diagonals import is_inactive_link
    >>> from landlab import CORE_NODE, FIXED_GRADIENT_BOUNDARY
    >>> is_inactive_link([CORE_NODE, CLOSED_BOUNDARY])
    array([ True], dtype=bool)

    >>> from landlab import FIXED_VALUE_BOUNDARY
    >>> is_inactive_link([FIXED_GRADIENT_BOUNDARY, FIXED_VALUE_BOUNDARY])
    array([ True], dtype=bool)

    >>> is_inactive_link([[FIXED_GRADIENT_BOUNDARY, CLOSED_BOUNDARY],
    ...                   [CORE_NODE, CORE_NODE]])
    array([ True, False], dtype=bool)
    """
    node_status_at_link = np.asarray(node_status_at_link).reshape((-1, 2))

    is_core = node_status_at_link == CORE_NODE
    is_fixed_value = node_status_at_link == FIXED_VALUE_BOUNDARY
    is_fixed_gradient = node_status_at_link == FIXED_GRADIENT_BOUNDARY
    is_closed = node_status_at_link == CLOSED_BOUNDARY
    is_boundary_node = is_fixed_value | is_fixed_gradient | is_closed

    return ((is_boundary_node[:, 0] & is_boundary_node[:, 1]) |
            (is_closed[:, 0] & is_core[:, 1]) |
            (is_core[:, 0] & is_closed[:, 1]))


def is_active_link(node_status_at_link):
    """Find links that are active.

    A link is active if it connects a core node with another core
    node or a fixed value boundary.

    Parameters
    ----------
    node_status_at_link : ndarray of int, shape `(n_links, 2)`
        Node status a link tail and head.

    Returns
    -------
    ndarray of bool, shape `(n_links, )`
        True if link is isactive.

    Examples
    --------
    >>> from landlab.grid.diagonals import is_active_link
    >>> from landlab import CORE_NODE, FIXED_GRADIENT_BOUNDARY
    >>> is_active_link([CORE_NODE, FIXED_GRADIENT_BOUNDARY])
    array([False], dtype=bool)

    >>> from landlab import FIXED_VALUE_BOUNDARY
    >>> is_active_link([CORE_NODE, FIXED_VALUE_BOUNDARY])
    array([ True], dtype=bool)

    >>> is_active_link([[FIXED_GRADIENT_BOUNDARY, CORE_NODE],
    ...                 [CORE_NODE, CORE_NODE]])
    array([False, True], dtype=bool)
    """
    node_status_at_link = np.asarray(node_status_at_link).reshape((-1, 2))

    is_core_node = node_status_at_link == CORE_NODE
    is_fixed_value_node = node_status_at_link == FIXED_VALUE_BOUNDARY
    return (
        (is_core_node[:, 0] & is_core_node[:, 1]) |
        (is_core_node[:, 0] & is_fixed_value_node[:, 1]) |
        (is_fixed_value_node[:, 0] & is_core_node[:, 1])
    )


def set_status_at_link(node_status_at_link, out=None):
    n_links = len(node_status_at_link)

    if out is None:
        out = np.full(n_links, 255, dtype=np.uint8)

    _is_fixed_link = is_fixed_link(node_status_at_link)
    _is_active_link = is_active_link(node_status_at_link)
    _is_inactive_link = is_inactive_link(node_status_at_link)

    assert np.all(np.sum(np.vstack((_is_active_link, _is_inactive_link,
                                    _is_fixed_link)), axis=0) == 1)

    out[_is_inactive_link] = INACTIVE_LINK
    out[_is_active_link] = ACTIVE_LINK
    out[_is_fixed_link] = FIXED_LINK

    return out





import numpy as np
cimport numpy as np
cimport cython


DTYPE = np.int
ctypedef np.int_t DTYPE_INT_t


@cython.boundscheck(False)
cpdef _add_to_stack(DTYPE_INT_t l, DTYPE_INT_t j,
                    np.ndarray[DTYPE_INT_t, ndim=1] s,
                    np.ndarray[DTYPE_INT_t, ndim=1] delta,
                    np.ndarray[DTYPE_INT_t, ndim=1] donors):

    """
    Adds node l to the stack and increments the current index (j).
    """
    cdef int m, n, delta_l, delta_lplus1

    s[j] = l
    j += 1
    delta_l = delta[l]
    delta_lplus1 = delta[l+1]

    for n in range(delta_l, delta_lplus1):
        m = donors[n]
        if m != l:
            j = _add_to_stack(m, j, s, delta, donors)

    return j




from landlab import Component
from ...utils.decorators import use_file_name_or_kwds
import numpy as np

_VALID_METHODS = set(['Grid'])


def assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)


class Vegetation(Component):
    """
    Landlab component that simulates net primary productivity, biomass
    and leaf area index at each cell based on inputs of root-zone
    average soil moisture.

    Ref: Zhou, X., Istanbulluoglu, E., & Vivoni, E. R. (2013). Modeling the
    ecohydrological role of aspect controlled radiation on tree grass shrub
    coexistence in a semiarid climate. Water Resources Research,
    49(5), 2872-2895.

    .. codeauthor:: Sai Nudurupati and Erkan Istanbulluoglu

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import Vegetation

    Create a grid on which to simulate vegetation dynamics.

    >>> grid = RasterModelGrid((5,4), spacing=(0.2, 0.2))

    The grid will need some input data. To check the names of the fields
    that provide the input to this component, use the *input_var_names*
    class property.

    >>> sorted(Vegetation.input_var_names)  # doctest: +NORMALIZE_WHITESPACE
    ['surface__evapotranspiration',
     'surface__potential_evapotranspiration_30day_mean',
     'surface__potential_evapotranspiration_rate',
     'vegetation__plant_functional_type',
     'vegetation__water_stress']

    >>> sorted(Vegetation.units) # doctest: +NORMALIZE_WHITESPACE
    [('surface__evapotranspiration', 'mm'),
     ('surface__potential_evapotranspiration_30day_mean', 'mm'),
     ('surface__potential_evapotranspiration_rate', 'mm'),
     ('vegetation__cover_fraction', 'None'),
     ('vegetation__dead_biomass', 'g m^-2 d^-1'),
     ('vegetation__dead_leaf_area_index', 'None'),
     ('vegetation__live_biomass', 'g m^-2 d^-1'),
     ('vegetation__live_leaf_area_index', 'None'),
     ('vegetation__plant_functional_type', 'None'),
     ('vegetation__water_stress', 'None')]

    Provide all the input fields.

    >>> grid['cell']['vegetation__plant_functional_type']= (
    ...         np.zeros(grid.number_of_cells, dtype=int))
    >>> grid['cell']['surface__evapotranspiration'] = (
    ...         0.2 * np.ones(grid.number_of_cells))
    >>> grid['cell']['surface__potential_evapotranspiration_rate']= np.array([
    ...         0.25547770, 0.25547770, 0.22110221,
    ...         0.22110221, 0.24813062, 0.24813062])
    >>> grid['cell']['surface__potential_evapotranspiration_30day_mean']= (
    ...        np.array([0.25547770, 0.25547770, 0.22110221,
    ...                  0.22110221, 0.24813062, 0.24813062]))
    >>> grid['cell']['vegetation__water_stress'] = (
    ...        0.01 * np.ones(grid.number_of_cells))

    Instantiate the 'Vegetation' component.

    >>> Veg = Vegetation(grid)

    >>> Veg.grid.number_of_cell_rows
    3
    >>> Veg.grid.number_of_cell_columns
    2
    >>> Veg.grid is grid
    True
    >>> import numpy as np
    >>> sorted(Vegetation.output_var_names)  # doctest: +NORMALIZE_WHITESPACE
    ['vegetation__cover_fraction',
     'vegetation__dead_biomass',
     'vegetation__dead_leaf_area_index',
     'vegetation__live_biomass',
     'vegetation__live_leaf_area_index']

    >>> np.all(grid.at_cell['vegetation__live_leaf_area_index'] == 0.)
    True

    >>> Veg.update()

    >>> np.all(grid.at_cell['vegetation__live_leaf_area_index'] == 0.)
    False
    """
    _name = 'Vegetation'

    _input_var_names = (
        'surface__evapotranspiration',
        'vegetation__water_stress',
        'surface__potential_evapotranspiration_rate',
        'surface__potential_evapotranspiration_30day_mean',
        'vegetation__plant_functional_type',
    )

    _output_var_names = (
        'vegetation__live_leaf_area_index',
        'vegetation__dead_leaf_area_index',
        'vegetation__cover_fraction',
        'vegetation__live_biomass',
        'vegetation__dead_biomass',
    )

    _var_units = {
        'vegetation__live_leaf_area_index': 'None',
        'vegetation__dead_leaf_area_index': 'None',
        'vegetation__cover_fraction': 'None',
        'surface__evapotranspiration': 'mm',
        'surface__potential_evapotranspiration_rate': 'mm',
        'surface__potential_evapotranspiration_30day_mean': 'mm',
        'vegetation__water_stress': 'None',
        'vegetation__live_biomass': 'g m^-2 d^-1',
        'vegetation__dead_biomass': 'g m^-2 d^-1',
        'vegetation__plant_functional_type': 'None',
    }

    _var_mapping = {
        'vegetation__live_leaf_area_index': 'cell',
        'vegetation__dead_leaf_area_index': 'cell',
        'vegetation__cover_fraction': 'cell',
        'surface__evapotranspiration': 'cell',
        'surface__potential_evapotranspiration_rate': 'cell',
        'surface__potential_evapotranspiration_30day_mean': 'cell',
        'vegetation__water_stress': 'cell',
        'vegetation__live_biomass': 'cell',
        'vegetation__dead_biomass': 'cell',
        'vegetation__plant_functional_type': 'cell',
    }

    _var_doc = {
        'vegetation__live_leaf_area_index':
            'one-sided green leaf area per unit ground surface area',
        'vegetation__dead_leaf_area_index':
            'one-sided dead leaf area per unit ground surface area',
        'vegetation__cover_fraction':
            'fraction of land covered by vegetation',
        'surface__evapotranspiration':
            'actual sum of evaporation and plant transpiration',
        'surface__potential_evapotranspiration_rate':
            'potential sum of evaporation and platnt transpiration',
        'surface__potential_evapotranspiration_30day_mean':
            '30 day mean of surface__potential_evapotranspiration',
        'vegetation__water_stress':
            'parameter that represents nonlinear effects of water defecit \
             on plants',
        'vegetation__live_biomass':
            'weight of green organic mass per unit area - measured in terms \
             of dry matter',
        'vegetation__dead_biomass':
            'weight of dead organic mass per unit area - measured in terms \
             of dry matter',
        'vegetation__plant_functional_type':
            'classification of plants (int), grass=0, shrub=1, tree=2, \
             bare=3, shrub_seedling=4, tree_seedling=5',
    }

    @use_file_name_or_kwds
    def __init__(self, grid, Blive_init=102., Bdead_init=450.,
                 ETthreshold_up=3.8, ETthreshold_down=6.8, Tdmax=10., w=0.55,
                 WUE_grass=0.01, LAI_max_grass=2., cb_grass=0.0047,
                 cd_grass=0.009, ksg_grass=0.012, kdd_grass=0.013,
                 kws_grass=0.02,
                 WUE_shrub=0.0025, LAI_max_shrub=2., cb_shrub=0.004,
                 cd_shrub=0.01, ksg_shrub=0.002, kdd_shrub=0.013,
                 kws_shrub=0.02,
                 WUE_tree=0.0045, LAI_max_tree=4., cb_tree=0.004, cd_tree=0.01,
                 ksg_tree=0.002, kdd_tree=0.013, kws_tree=0.01,
                 WUE_bare=0.01, LAI_max_bare=0.01, cb_bare=0.0047,
                 cd_bare=0.009, ksg_bare=0.012, kdd_bare=0.013, kws_bare=0.02,
                 **kwds):
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A grid.
        Blive_init: float, optional
            Initial value for vegetation__live_biomass. Converted to field.
        Bdead_init: float, optional
            Initial value for vegetation__dead_biomass. Coverted to field.
        ETthreshold_up: float, optional
            Potential Evapotranspiration (PET) threshold for
            growing season (mm/d).
        ETthreshold_down: float, optional
            PET threshold for dormant season (mm/d).
        Tdmax: float, optional
            Constant for dead biomass loss adjustment (mm/d).
        w: float, optional
            Conversion factor of CO2 to dry biomass (Kg DM/Kg CO2).
        WUE: float, optional
            Water Use Efficiency - ratio of water used in plant water
            lost by the plant through transpiration (KgCO2Kg-1H2O).
        LAI_max: float, optional
            Maximum leaf area index (m2/m2).
        cb: float, optional
            Specific leaf area for green/live biomass (m2 leaf g-1 DM).
        cd: float, optional
            Specific leaf area for dead biomass (m2 leaf g-1 DM).
        ksg: float, optional
            Senescence coefficient of green/live biomass (d-1).
        kdd: float, optional
            Decay coefficient of aboveground dead biomass (d-1).
        kws: float, optional
            Maximum drought induced foliage loss rate (d-1).
        """
        self._method = kwds.pop('method', 'Grid')

        assert_method_is_valid(self._method)

        super(Vegetation, self).__init__(grid)

        self.initialize(Blive_init=Blive_init, Bdead_init=Bdead_init,
                        ETthreshold_up=ETthreshold_up,
                        ETthreshold_down=ETthreshold_down, Tdmax=Tdmax, w=w,
                        WUE_grass=WUE_grass, LAI_max_grass=LAI_max_grass,
                        cb_grass=cb_grass, cd_grass=cd_grass,
                        ksg_grass=ksg_grass, kdd_grass=kdd_grass,
                        kws_grass=kws_grass, WUE_shrub=WUE_shrub,
                        LAI_max_shrub=LAI_max_shrub, cb_shrub=cb_shrub,
                        cd_shrub=cd_shrub, ksg_shrub=ksg_shrub,
                        kdd_shrub=kdd_shrub, kws_shrub=kws_shrub,
                        WUE_tree=WUE_tree, LAI_max_tree=LAI_max_tree,
                        cb_tree=cb_tree, cd_tree=cd_tree, ksg_tree=ksg_tree,
                        kdd_tree=kdd_tree, kws_tree=kws_tree,
                        WUE_bare=WUE_bare, LAI_max_bare=LAI_max_bare,
                        cb_bare=cb_bare, cd_bare=cd_bare, ksg_bare=ksg_bare,
                        kdd_bare=kdd_bare, kws_bare=kws_bare, **kwds)

        for name in self._input_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        for name in self._output_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        self._cell_values = self.grid['cell']

        self._Blive_ini = self._Blive_init * np.ones(self.grid.number_of_cells)
        self._Bdead_ini = self._Bdead_init * np.ones(self.grid.number_of_cells)

    def initialize(self, Blive_init=102., Bdead_init=450., ETthreshold_up=3.8,
                   ETthreshold_down=6.8, Tdmax=10., w=0.55,
                   WUE_grass=0.01, LAI_max_grass=2., cb_grass=0.0047,
                   cd_grass=0.009, ksg_grass=0.012, kdd_grass=0.013,
                   kws_grass=0.02,
                   WUE_shrub=0.0025, LAI_max_shrub=2., cb_shrub=0.004,
                   cd_shrub=0.01, ksg_shrub=0.002, kdd_shrub=0.013,
                   kws_shrub=0.02,
                   WUE_tree=0.0045, LAI_max_tree=4., cb_tree=0.004,
                   cd_tree=0.01, ksg_tree=0.002, kdd_tree=0.013, kws_tree=0.01,
                   WUE_bare=0.01, LAI_max_bare=0.01, cb_bare=0.0047,
                   cd_bare=0.009, ksg_bare=0.012, kdd_bare=0.013,
                   kws_bare=0.02, **kwds):
        # GRASS = 0; SHRUB = 1; TREE = 2; BARE = 3;
        # SHRUBSEEDLING = 4; TREESEEDLING = 5
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A grid.
        Blive_init: float, optional
            Initial value for vegetation__live_biomass. Converted to field.
        Bdead_init: float, optional
            Initial value for vegetation__dead_biomass. Coverted to field.
        ETthreshold_up: float, optional
            Potential Evapotranspiration (PET) threshold for
            growing season (mm/d).
        ETthreshold_down: float, optional
            PET threshold for dormant season (mm/d).
        Tdmax: float, optional
            Constant for dead biomass loss adjustment (mm/d).
        w: float, optional
            Conversion factor of CO2 to dry biomass (Kg DM/Kg CO2).
        WUE: float, optional
            Water Use Efficiency - ratio of water used in plant water
            lost by the plant through transpiration (KgCO2Kg-1H2O).
        LAI_max: float, optional
            Maximum leaf area index (m2/m2).
        cb: float, optional
            Specific leaf area for green/live biomass (m2 leaf g-1 DM).
        cd: float, optional
            Specific leaf area for dead biomass (m2 leaf g-1 DM).
        ksg: float, optional
            Senescence coefficient of green/live biomass (d-1).
        kdd: float, optional
            Decay coefficient of aboveground dead biomass (d-1).
        kws: float, optional
            Maximum drought induced foliage loss rate (d-1).
        """
        self._vegtype = self.grid['cell']['vegetation__plant_functional_type']
        self._WUE = np.choose(self._vegtype, [
            WUE_grass, WUE_shrub, WUE_tree, WUE_bare, WUE_shrub, WUE_tree])
        # Water Use Efficiency  KgCO2kg-1H2O
        self._LAI_max = np.choose(self._vegtype, [
            LAI_max_grass, LAI_max_shrub, LAI_max_tree,
            LAI_max_bare, LAI_max_shrub, LAI_max_tree])
        # Maximum leaf area index (m2/m2)
        self._cb = np.choose(self._vegtype, [
            cb_grass, cb_shrub, cb_tree, cb_bare, cb_shrub, cb_tree])
        # Specific leaf area for green/live biomass (m2 leaf g-1 DM)
        self._cd = np.choose(self._vegtype, [
            cd_grass, cd_shrub, cd_tree, cd_bare, cd_shrub, cd_tree])
        # Specific leaf area for dead biomass (m2 leaf g-1 DM)
        self._ksg = np.choose(self._vegtype, [
            ksg_grass, ksg_shrub, ksg_tree, ksg_bare, ksg_shrub, ksg_tree])
        # Senescence coefficient of green/live biomass (d-1)
        self._kdd = np.choose(self._vegtype, [
            kdd_grass, kdd_shrub, kdd_tree, kdd_bare, kdd_shrub, kdd_tree])
        # Decay coefficient of aboveground dead biomass (d-1)
        self._kws = np.choose(self._vegtype, [
            kws_grass, kws_shrub, kws_tree, kws_bare, kws_shrub, kws_tree])
        # Maximum drought induced foliage loss rates (d-1)
        self._Blive_init = Blive_init
        self._Bdead_init = Bdead_init
        self._ETthresholdup = ETthreshold_up    # Growth threshold (mm/d)
        self._ETthresholddown = ETthreshold_down   # Dormancy threshold (mm/d)
        self._Tdmax = Tdmax         # Constant for dead biomass loss adjustment
        self._w = w                  # Conversion factor of CO2 to dry biomass

        self._Blive_ini = self._Blive_init * np.ones(self.grid.number_of_cells)
        self._Bdead_ini = self._Bdead_init * np.ones(self.grid.number_of_cells)

    def update(self, PETthreshold_switch=0, Tb=24., Tr=0.01, **kwds):
        """
        Update fields with current loading conditions.

        Parameters
        ----------
        Tr: float, optional
            Storm duration (hours).
        Tb: float, optional
            Inter-storm duration (hours).
        """
        PETthreshold_ = PETthreshold_switch
        PET = self._cell_values['surface__potential_evapotranspiration_rate']
        PET30_ = self._cell_values[
            'surface__potential_evapotranspiration_30day_mean']
        ActualET = self._cell_values['surface__evapotranspiration']
        Water_stress = self._cell_values['vegetation__water_stress']

        self._LAIlive = self._cell_values['vegetation__live_leaf_area_index']
        self._LAIdead = self._cell_values['vegetation__dead_leaf_area_index']
        self._Blive = self._cell_values['vegetation__live_biomass']
        self._Bdead = self._cell_values['vegetation__dead_biomass']
        self._VegCov = self._cell_values['vegetation__cover_fraction']

        if PETthreshold_ == 1:
            PETthreshold = self._ETthresholdup
        else:
            PETthreshold = self._ETthresholddown

        for cell in range(0, self.grid.number_of_cells):

            WUE = self._WUE[cell]
            LAImax = self._LAI_max[cell]
            cb = self._cb[cell]
            cd = self._cd[cell]
            ksg = self._ksg[cell]
            kdd = self._kdd[cell]
            kws = self._kws[cell]
            # ETdmax = self._ETdmax[cell]
            LAIlive = min(cb*self._Blive_ini[cell], LAImax)
            LAIdead = min(cd * self._Bdead_ini[cell], (LAImax -
                          LAIlive))
            NPP = max((ActualET[cell]/(Tb+Tr)) *
                      WUE*24.*self._w*1000, 0.001)

            if self._vegtype[cell] == 0:
                if PET30_[cell] > PETthreshold:
                                # Growing Season
                    Bmax = (LAImax - LAIdead)/cb
                    Yconst = (1/((1/Bmax)+(((kws*Water_stress[cell]) +
                              ksg)/NPP)))
                    Blive = ((self._Blive_ini[cell] - Yconst) *
                             np.exp(-(NPP/Yconst) * ((Tb+Tr)/24.)) + Yconst)
                    Bdead = ((self._Bdead_ini[cell] + (Blive - max(Blive *
                             np.exp(-1 * ksg * Tb/24.), 0.00001))) *
                             np.exp(-1 * kdd *
                             min(PET[cell]/self._Tdmax, 1.) * Tb/24.))
                else:                                 # Senescense
                    Blive = max(self._Blive_ini[cell] * np.exp((-2) * ksg *
                                Tb/24.), 1)
                    Bdead = max((self._Bdead_ini[cell]+(self._Blive_ini[cell] -
                                (max(self._Blive_ini[cell]*np.exp((-2) *
                                 ksg*Tb/24.), 0.000001)))*np.exp((-1)*kdd *
                                 min(PET[cell]/self._Tdmax, 1.) * Tb/24.), 0.))

            elif self._vegtype[cell] == 3:
                Blive = 0.
                Bdead = 0.

            else:
                Bmax = LAImax/cb
                Yconst = (1./((1./Bmax)+(((kws*Water_stress[cell]) +
                          ksg)/NPP)))
                Blive = ((self._Blive_ini[cell] - Yconst) *
                         np.exp(-(NPP/Yconst) * ((Tb+Tr)/24.)) + Yconst)
                Bdead = ((self._Bdead_ini[cell] + (Blive - max(Blive *
                         np.exp(-ksg * Tb/24.), 0.00001))) *
                         np.exp(-kdd * min(PET[cell]/self._Tdmax, 1.) *
                         Tb/24.))

            LAIlive = min(cb * (Blive + self._Blive_ini[cell])/2., LAImax)
            LAIdead = min(cd * (Bdead + self._Bdead_ini[cell])/2.,
                          (LAImax - LAIlive))
            if self._vegtype[cell] == 0:
                Vt = 1. - np.exp(-0.75 * (LAIlive + LAIdead))
            else:
                # Vt = 1 - np.exp(-0.75 * LAIlive)
                Vt = 1.

            self._LAIlive[cell] = LAIlive
            self._LAIdead[cell] = LAIdead
            self._VegCov[cell] = Vt
            self._Blive[cell] = Blive
            self._Bdead[cell] = Bdead

        self._Blive_ini = self._Blive
        self._Bdead_ini = self._Bdead



"""
Unit tests for \
    landlab.components.vegetation_dynamics.vegetation_dynamics
"""
import pytest
from numpy.testing import assert_array_almost_equal
import numpy as np


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(veg):
    assert veg.name == 'Vegetation'


def test_input_var_names(veg):
    assert sorted(veg.input_var_names) == ['surface__evapotranspiration',
                                           'surface__potential_evapotranspiration_30day_mean',
                                           'surface__potential_evapotranspiration_rate',
                                           'vegetation__plant_functional_type',
                                           'vegetation__water_stress']


def test_output_var_names(veg):
    assert sorted(veg.output_var_names) == ['vegetation__cover_fraction',
                                            'vegetation__dead_biomass',
                                            'vegetation__dead_leaf_area_index',
                                            'vegetation__live_biomass',
                                            'vegetation__live_leaf_area_index']


def test_var_units(veg):
    assert set(veg.input_var_names) | set(veg.output_var_names) == set(dict(veg.units).keys())

    assert veg.var_units('vegetation__live_leaf_area_index') == 'None'
    assert veg.var_units('vegetation__dead_leaf_area_index') == 'None'
    assert veg.var_units('vegetation__cover_fraction') == 'None'
    assert veg.var_units('surface__evapotranspiration') == 'mm'
    assert veg.var_units('surface__potential_evapotranspiration_rate') == 'mm'
    assert veg.var_units(
                'surface__potential_evapotranspiration_30day_mean') == 'mm'
    assert veg.var_units('vegetation__water_stress') == 'None'
    assert veg.var_units('vegetation__live_biomass') == 'g m^-2 d^-1'
    assert veg.var_units('vegetation__dead_biomass') == 'g m^-2 d^-1'
    assert veg.var_units('vegetation__plant_functional_type') == 'None'


def test_grid_shape(veg):
    assert veg.grid.number_of_node_rows == _SHAPE[0]
    assert veg.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(veg):
    assert veg.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(veg):
    assert veg.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(veg):
    for name in veg.grid['node']:
        field = veg.grid['node'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (veg.grid.number_of_node_rows * veg.grid.number_of_node_columns, )
                      
    for name in veg.grid['cell']:
        field = veg.grid['cell'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (veg.grid.number_of_cell_rows * veg.grid.number_of_cell_columns, )

    with pytest.raises(KeyError):
        veg.grid['not_a_var_name']


def test_field_initialized_to_zero(veg):
    for name in veg.grid['node']:
        field = veg.grid['node'][name]
        assert_array_almost_equal(field, np.zeros(veg.grid.number_of_nodes))
    for name in veg.grid['cell']:
        field = veg.grid['cell'][name]
        assert_array_almost_equal(field, np.zeros(veg.grid.number_of_cells))



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.grid.raster_gradients import calc_grad_across_cell_corners


def test_scalar_arg():
    """Test using a scalar for cell arg."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    grads = calc_grad_across_cell_corners(
        rmg, values_at_nodes, 0)
    assert_array_equal(grads, np.array([[6., 4., -6., -4.]]) / np.sqrt(2.))


def test_iterable():
    """Test using an iterable for cell arg."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    grads = rmg.calc_grad_across_cell_corners(values_at_nodes, [0, 4])
    assert_array_equal(grads, np.array([[6., 4., -6., -4.],
                                        [6., 4., -6., -4.]]) / np.sqrt(2.))


def test_with_no_cell_id_arg():
    """Test without using an arg for cell id."""
    rmg = RasterModelGrid((4, 5))
    values = np.array([0, 1,  3, 6, 10,
                       0, 1,  3, 6, 10,
                       0, 1,  3, 5, 10,
                       0, 1, -3, 6, 10], dtype=float)
    grads = rmg.calc_grad_across_cell_corners(values)

    assert_array_equal(grads, (1. / np.sqrt(2.)) * np.array([
        [2., -1., -1., 2.], [2., -2., -2., 3.], [4., -3., -3., 4.],
        [-4., -1., -1., 2.], [3., -2., -2., 3.], [5., -8., -2., 5.]]))


def test_with_out_keyword():
    """Test with out keyword."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    out = np.empty((1, 4))
    rtn = rmg.calc_grad_across_cell_corners(values_at_nodes, 5, out=out)
    assert rtn is out
    assert_array_equal(out, np.array([[6., 4., -6., -4.]]) / np.sqrt(2))



"""
Creates a field of Manning's n value, where each value is dependent on the
water depth at a given node (e.g. Jain et al., 2005, Mugler et al., 2011 and
Rengers et al., 2016). This "effectively creates separate values of Manning's n
for hillslopes and channels" (Rengers et al., 2016).

This component creates -or- overwrites a field on the grid called
'mannings_n' where each node is assigned a Manning's n value based on the
minimum Manning's n value for the landscape, the local water depths, an index
(or threshold) water depth above which all Manning's n values are considered
constant, and a vegetation drag coefficent (for more on vegetation drag and
the impact on surface roughness, see Wu et al., 1999 in the Journal of
Hydraulic Engineering.)

This can be used iteratively inside a driver loop, to update Manning's n values
as water depths change in another component (e.g. OverlandFlow)

.. codeauthor:: Jordan Adams

Examples
--------
>>> import numpy as np
>>> from landlab import RasterModelGrid
>>> grid = RasterModelGrid((5, 5))
>>> grid.at_node['surface_water__depth'] = np.array(
...     [5., 5., 5., 5., 5.,
...     4., 4., 4., 4., 4.,
...     3., 3., 3., 3., 3.,
...     2., 2., 2., 2., 2.,
...     1., 1., 1., 1., 1.])
>>> depth_dependent_mannings_n(grid, index_flow_depth=2.)
>>> grid.at_node['mannings_n']
array([ 0.06      ,  0.06      ,  0.06      ,  0.06      ,  0.06      ,
        0.06      ,  0.06      ,  0.06      ,  0.06      ,  0.06      ,
        0.06      ,  0.06      ,  0.06      ,  0.06      ,  0.06      ,
        0.06      ,  0.06      ,  0.06      ,  0.06      ,  0.06      ,
        0.07559526,  0.07559526,  0.07559526,  0.07559526,  0.07559526])
"""

import numpy as np
from landlab import FieldError


def depth_dependent_mannings_n(grid, water_depths='surface_water__depth',
                               min_mannings_n=0.06,
                               index_flow_depth=0.003,
                               veg_drag_exponent=(-1./3.)):
    """
    Method to create or overwrite a Manning's n field
    (grid.at_node['mannings_n']) with spatially variable n values based on
    water depths.

    Parameters
    ----------
    grid : A Landlab RasterModelGrid instance
        A Landlab grid - only works with RasterModelGrid instances as of
        1/31/17.
    water_depths : array or Landlab field of floats
        Array of values, with length of number of nodes, water depths
        at all grid node locations. (m)
    min_mannings_n : float
        This is the minimum Manning's n coefficient for a given landscape,
        following Chow, 1959. (s m^(-1./3.))
    index_flow_depth : float
        The flow depth above which it is assumed that Manning's n is
        constant. (m)
    veg_drag_exponent :
        An exponent related to vegetation drag effects, which increases
        effective Manning's n at low flow conditions.
    """

    ## Looks for a field called 'mannings_n' attached to the grid instance. If
    ## one is found, a FieldError is thrown but ignored. This method
    ## REWRITES over the existing Manning's n fields after the calculation.
    try:
        grid.add_zeros('mannings_n', at='node')

    except FieldError:
        pass

    ## Identifies locations where water depth is lower than the value supplied
    ## through keyword index_flow_depth.
    (locs_less, ) = np.where(grid.at_node['surface_water__depth'] <=
                             index_flow_depth)

    ## Identifies locations where water depth is greater than the value
    ## supplied through keyword index_flow_depth.
    (locs_more, ) = np.where(grid.at_node['surface_water__depth'] >
                             index_flow_depth)

    ## At all locations lower than the index flow depth (assumed to be shallow
    ## flow on hillslopes), a new Manning's n value is calculated to that
    ## incorporates effects of vegetation drag. These Manning's n values will
    ## be greater than the supplied Manning's n (keyword: min_mannings_n)
    grid.at_node['mannings_n'][locs_less] = (
        min_mannings_n * (grid.at_node['surface_water__depth'][locs_less] /
                          index_flow_depth) ** veg_drag_exponent)

    ## Resets the field with the new Manning's n values.
    grid.at_node['mannings_n'][locs_more] = min_mannings_n



import pytest

from landlab import RasterModelGrid
from landlab.components.soil_moisture.soil_moisture_dynamics import SoilMoisture
from landlab.components.soil_moisture import SoilInfiltrationGreenAmpt


@pytest.fixture
def sm():
    grid = RasterModelGrid((20, 20), spacing=10e0)
    grid.add_zeros('vegetation__plant_functional_type', at='cell', dtype=int)
    return SoilMoisture(grid)


@pytest.fixture
def si():
    grid = RasterModelGrid((10, 10), spacing=25)
    grid.add_ones('soil_water_infiltration__depth', at='node', dtype=float)
    grid.add_ones('surface_water__depth', at='node')
    hydraulic_conductivity = (2.5 *  (10**-5))
    grid.at_node['surface_water__depth'] *= 0.5
    grid.at_node['soil_water_infiltration__depth'] *= (10**-5)
    return SoilInfiltrationGreenAmpt(
        grid,
        hydraulic_conductivity=hydraulic_conductivity,
        soil_bulk_density=1700., rock_density=2650.,
        initial_soil_moisture_content=0.2, soil_type='silt loam',
        volume_fraction_coarse_fragments=0.6,
        coarse_sed_flag=False,
        surface_water_minimum_depth=1.e-7,
        soil_pore_size_distribution_index=None,
        soil_bubbling_pressure=None,
        wetting_front_capillary_pressure_head=None)



"""
Unit tests for landlab.components.plant_competition_ca.plant_competition_ca
"""
import pytest
from numpy.testing import assert_array_almost_equal
import numpy as np


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(ca_veg):
    assert ca_veg.name == 'Cellular Automata Plant Competition'


def test_input_var_names(ca_veg):
    assert sorted(ca_veg.input_var_names) == [
        'vegetation__cumulative_water_stress',
        'vegetation__plant_functional_type',
    ]


def test_output_var_names(ca_veg):
    assert sorted(ca_veg.output_var_names) == [
        'plant__age',
        'plant__live_index',
    ]


def test_var_units(ca_veg):
    assert set(ca_veg.input_var_names) | set(ca_veg.output_var_names) == set(dict(ca_veg.units).keys())

    assert ca_veg.var_units('vegetation__cumulative_water_stress') == 'None'
    assert ca_veg.var_units('vegetation__plant_functional_type') == 'None'
    assert ca_veg.var_units('plant__live_index') == 'None'
    assert ca_veg.var_units('plant__age') == 'Years'


def test_grid_shape(ca_veg):
    assert ca_veg.grid.number_of_node_rows == _SHAPE[0]
    assert ca_veg.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(ca_veg):
    assert ca_veg.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(ca_veg):
    assert ca_veg.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(ca_veg):
    for name in ca_veg.grid['node']:
        field = ca_veg.grid['node'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (ca_veg.grid.number_of_node_rows *
                               ca_veg.grid.number_of_node_columns, )
                      
    for name in ca_veg.grid['cell']:
        field = ca_veg.grid['cell'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape, (ca_veg.grid.number_of_cell_rows *
                             ca_veg.grid.number_of_cell_columns, )

    with pytest.raises(KeyError):
        ca_veg.grid['not_a_var_name']


def test_field_initialized_to_zero(ca_veg):
    for name in ca_veg.grid['node']:
        field = ca_veg.grid['node'][name]
        assert_array_almost_equal(field, np.zeros(ca_veg.grid.number_of_nodes))



#!/usr/env/python
"""
Base class for a "full" generic CTS model.
"""

_DEBUG = False

import time
#import random
from numpy import zeros, bincount, arange, savetxt, sqrt, log10, mean, arctan, pi, random
from landlab.io.native_landlab import save_grid
from landlab.ca.celllab_cts import Transition, CAPlotter
from matplotlib.pyplot import axis
import sys


class CTSModel(object):
    """
    Implement a generic CellLab-CTS model.

    This is the base class from which models should inherit.
    """

    def __init__(self, grid_size=(5, 5), report_interval=5.0,
                 grid_orientation='vertical', grid_shape='rect',
                 show_plots=False, cts_type='oriented_hex', 
                 run_duration=1.0, output_interval=1.0e99,
                 plot_every_transition=False, **kwds):

        self.initialize(grid_size, report_interval, grid_orientation,
                        grid_shape, show_plots, cts_type, run_duration,
                        output_interval, plot_every_transition, **kwds)


    def initialize(self, grid_size=(5, 5), report_interval=5.0,
                 grid_orientation='vertical', grid_shape='rect',
                 show_plots=False, cts_type='oriented_hex', 
                 run_duration=1.0, output_interval=1.0e99,
                 plot_every_transition=False, **kwds):

        # Remember the clock time, and calculate when we next want to report
        # progress.
        self.current_real_time = time.time()
        self.next_report = self.current_real_time + report_interval
        self.report_interval = report_interval

        # Interval for output
        self.output_interval = output_interval

        # Duration for run
        self.run_duration = run_duration

        # Create a grid
        self.create_grid_and_node_state_field(grid_size[0], grid_size[1], 
                                              grid_orientation, grid_shape,
                                              cts_type)

        # Create the node-state dictionary
        ns_dict = self.node_state_dictionary()

        # Initialize values of the node-state grid
        nsg = self.initialize_node_state_grid()
        
        # Create the transition list
        xn_list = self.transition_list()

        # Create the CA object
        if cts_type == 'raster':
            from landlab.ca.raster_cts import RasterCTS
            self.ca = RasterCTS(self.grid, ns_dict, xn_list, nsg)
        elif cts_type == 'oriented_raster':
            from landlab.ca.oriented_raster_cts import OrientedRasterCTS
            self.ca = OrientedRasterCTS(self.grid, ns_dict, xn_list, nsg)
        elif cts_type == 'hex':
            from landlab.ca.hex_cts import HexCTS
            self.ca = HexCTS(self.grid, ns_dict, xn_list, nsg)
        else:
            from landlab.ca.oriented_hex_cts import OrientedHexCTS
            self.ca = OrientedHexCTS(self.grid, ns_dict, xn_list, nsg)

        # Initialize graphics
        self._show_plots = show_plots
        if show_plots == True:
            self.initialize_plotting(**kwds)


    def create_grid_and_node_state_field(self, num_rows, num_cols, 
                                         grid_orientation, grid_shape,
                                         cts_type):
        """Create the grid and the field containing node states."""

        if cts_type == 'raster' or cts_type == 'oriented_raster':
            from landlab import RasterModelGrid
            self.grid = RasterModelGrid(shape=(num_rows, num_cols),
                                        spacing=1.0)
        else:
            from landlab import HexModelGrid
            self.grid = HexModelGrid(num_rows, num_cols, 1.0, 
                                     orientation=grid_orientation, 
                                     shape=grid_shape)

        self.grid.add_zeros('node', 'node_state', dtype=int)


    def node_state_dictionary(self):
        """Create and return a dictionary of all possible node (cell) states.
        
        This method creates a default set of states (just two); it is a
        template meant to be overridden.
        """
        ns_dict = { 0 : 'on', 
                    1 : 'off'}
        return ns_dict


    def transition_list(self):
        """Create and return a list of transition objects.
        
        This method creates a default set of transitions (just two); it is a
        template meant to be overridden.
        """
        xn_list = []
        xn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
        xn_list.append(Transition((1, 0, 0), (0, 1, 0), 1.0))
        return xn_list


    def write_output(self, grid, outfilename, iteration):
        """Write output to file (currently netCDF)."""
        filename = outfilename + str(iteration).zfill(4) + '.nc'
        save_grid(grid, filename)


    def initialize_node_state_grid(self):
        """Initialize values in the node-state grid.
        
        This method should be overridden. The default is random "on" and "off".        
        """
        num_states = 2
        for i in range(self.grid.number_of_nodes):
            self.grid.at_node['node_state'][i] = random.randint(num_states)
        return self.grid.at_node['node_state']


    def initialize_plotting(self, **kwds):
        """Create and configure CAPlotter object."""
        self.ca_plotter = CAPlotter(self.ca, **kwds)
        self.ca_plotter.update_plot()
        axis('off')


    def run_for(self, dt):

        self.ca.run(self.ca.current_time + dt, self.ca.node_state)


if __name__ == '__main__':
    ctsm = CTSModel(show_plots=True)
    ctsm.run_for(1.0)
    ctsm.ca_plotter.update_plot()



import os
import re

import six
import yaml

from .model_parameter_dictionary import ModelParameterDictionary


_loader = yaml.SafeLoader
_loader.add_implicit_resolver(
    u'tag:yaml.org,2002:float',
    re.compile(u'''^(?:
               [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
               |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
               |\\.[0-9_]+(?:[eE][-+][0-9]+)?
               |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
               |[-+]?\\.(?:inf|Inf|INF)
               |\\.(?:nan|NaN|NAN))$''', re.X),
    list(u'-+0123456789.'))


def load_file_contents(file_like):
    """Load the contents of a file or file-like object.

    Parameters
    ----------
    file_like : file_like or str
        File to load either as a file-like object, path to an existing file,
        or the contents of a file.

    Returns
    -------
    str
        The contents of the file.
    """
    try:
        contents = file_like.read()
    except AttributeError:  # was a str
        if os.path.isfile(file_like):
            with open(file_like, 'r') as fp:
                contents = fp.read()
        else:
            contents = file_like

    return contents


def load_params(file_like):
    """Load parameters from a file.

    Parameters
    ----------
    file_like : file_like or str
        Contents of a parameter file, a file-like object, or the path to
        a parameter file.

    Returns
    -------
    dict
        Parameters as key-value pairs.

    Examples
    --------
    >>> from landlab.core import load_params
    >>> contents = \"\"\"
    ... start: 0.
    ... stop: 10.
    ... step: 2.
    ... \"\"\"
    >>> params = load_params(contents)
    >>> isinstance(params, dict)
    True
    >>> params['start'], params['stop'], params['step']
    (0.0, 10.0, 2.0)

    >>> contents = \"\"\"
    ... start: Start time
    ... 0.
    ... stop: Stop time
    ... 10.
    ... step: Step time
    ... 2.
    ... \"\"\"
    >>> params = load_params(contents)
    >>> isinstance(params, dict)
    True
    >>> params['start'], params['stop'], params['step']
    (0.0, 10.0, 2.0)
    """
    contents = load_file_contents(file_like)

    try:
        params = yaml.load(contents, Loader=_loader)
    except yaml.YAMLError:
        file_like = six.StringIO(contents)
        params = ModelParameterDictionary(from_file=file_like, auto_type=True)

    if not isinstance(params, dict):
        raise ValueError(
            'parsing of parameter file did not produce a dict-like object')

    return params



# -*- coding: utf-8 -*-
"""
Created on Mon Jul 11 10:00:25 2016

@author: Charlie Shobe

Landlab component to calculate drainage density

"""
from warnings import warn
import numpy as np

from landlab import Component
from landlab import FieldError


class DrainageDensity(Component):

    """Calculate drainage density over a DEM.

    calc_drainage_density function returns drainage density for the model
    domain.

    Landlab component that implements the distance to channel algorithm of
    Tucker et al., 2001.

    This component requires EITHER a channel__mask array with 1's
    where channels exist and 0's elsewhere, OR a set of coefficients
    and exponents for a slope-area relationship and a
    channelization threshold to compare against that relationship.

    Written by C. Shobe on 7/11/2016, modified 10/7/2016.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.flow_routing import FlowRouter
    >>> from landlab.components import FastscapeEroder
    >>> mg = RasterModelGrid((10, 10), 1.0)
    >>> _ = mg.add_zeros('node', 'topographic__elevation')
    >>> np.random.seed(50)
    >>> noise = np.random.rand(100)
    >>> mg.at_node['topographic__elevation'] += noise
    >>> mg.at_node['topographic__elevation'] # doctest: +NORMALIZE_WHITESPACE
    array([ 0.49460165,  0.2280831 ,  0.25547392,  0.39632991,  0.3773151 ,
        0.99657423,  0.4081972 ,  0.77189399,  0.76053669,  0.31000935,
        0.3465412 ,  0.35176482,  0.14546686,  0.97266468,  0.90917844,
        0.5599571 ,  0.31359075,  0.88820004,  0.67457307,  0.39108745,
        0.50718412,  0.5241035 ,  0.92800093,  0.57137307,  0.66833757,
        0.05225869,  0.3270573 ,  0.05640164,  0.17982769,  0.92593317,
        0.93801522,  0.71409271,  0.73268761,  0.46174768,  0.93132927,
        0.40642024,  0.68320577,  0.64991587,  0.59876518,  0.22203939,
        0.68235717,  0.8780563 ,  0.79671726,  0.43200225,  0.91787822,
        0.78183368,  0.72575028,  0.12485469,  0.91630845,  0.38771099,
        0.29492955,  0.61673141,  0.46784623,  0.25533891,  0.83899589,
        0.1786192 ,  0.22711417,  0.65987645,  0.47911625,  0.07344734,
        0.13896007,  0.11230718,  0.47778497,  0.54029623,  0.95807105,
        0.58379231,  0.52666409,  0.92226269,  0.91925702,  0.25200886,
        0.68263261,  0.96427612,  0.22696165,  0.7160172 ,  0.79776011,
        0.9367512 ,  0.8537225 ,  0.42154581,  0.00543987,  0.03486533,
        0.01390537,  0.58890993,  0.3829931 ,  0.11481895,  0.86445401,
        0.82165703,  0.73749168,  0.84034417,  0.4015291 ,  0.74862   ,
        0.55962945,  0.61323757,  0.29810165,  0.60237917,  0.42567684,
        0.53854438,  0.48672986,  0.49989164,  0.91745948,  0.26287702])
    >>> fr = FlowRouter(mg)
    >>> fsc = FastscapeEroder(mg, K_sp=.01, m_sp=.5, n_sp=1)
    >>> for x in range(100):
    ...     fr.run_one_step()
    ...     fsc.run_one_step(dt = 10.0)
    ...     mg.at_node['topographic__elevation'][mg.core_nodes] += .01
    >>> channels = mg.at_node['drainage_area'] > 5
    >>> dd = DrainageDensity(mg, channel__mask=channels)
    >>> mean_drainage_density = dd.calc_drainage_density()
    >>> np.isclose(mean_drainage_density, 0.3831100571)
    True
    """

    _name = 'DrainageDensity'

    _input_var_names = (
        'flow__receiver_node',
        'flow__link_to_receiver_node',
        'topographic__steepest_slope',
        'channel__mask',
        'area_coefficient',
        'slope_coefficient',
        'area_exponent',
        'slope_exponent',
        'channelization_threshold',
    )

    _output_var_names = (
        'surface_to_channel__minimum_distance',
    )

    _var_units = {
        'flow__receiver_node': '-',
        'flow__link_to_receiver_node': '-',
        'topographic__steepest_slope': '-',
        'channel__mask': '-',
        'surface_to_channel__minimum_distance': 'm',
    }

    _var_mapping = {
        'flow__receiver_node': 'node',
        'flow__link_to_receiver_node': 'node',
        'topographic__steepest_slope': 'node',
        'channel__mask': 'node',
        'surface_to_channel__minimum_distance': 'node',
    }

    _var_doc = {
        'flow__receiver_node':
            'Node array of receivers (node that receives flow from current '
            'node)',
        'flow__link_to_receiver_node':
            'ID of link downstream of each node, which carries the discharge',
        'topographic__steepest_slope':
            'Topographic slope at each node',
        'channel__mask':
            'Logical map of at which grid nodes channels are present',
        'surface_to_channel__minimum_distance':
            'Distance from each node to the nearest channel',
    }

    def __init__(self, grid, channel__mask=None, area_coefficient=None,
                 slope_coefficient=None, area_exponent=None,
                 slope_exponent=None, channelization_threshold=None,
                 **kwds):
        """Initialize the DrainageDensity component.

        Parameters
        ----------
        grid : ModelGrid
        channel__mask : Array that holds 1's where
            channels exist and 0's elsewhere
        area_coefficient : coefficient to multiply drainage area by,
            for calculating channelization threshold
        slope_coefficient : coefficient to multiply slope by,
            for calculating channelization threshold
        area_exponent : exponent to raise drainage area to,
            for calculating channelization threshold
        slope_exponent : exponent to raise slope to,
            for calculating channelization threshold
        channelization_threshold : threshold value above
            which channels exist
        """

        if channel__mask is not None:
            if area_coefficient is not None:
                warn('Channel mask and area '
                     'coefficient supplied. Defaulting '
                     'to channel mask, ignoring area '
                     'coefficient.')
            if slope_coefficient is not None:
                warn('Channel mask and slope '
                     'coefficient supplied. Defaulting '
                     'to channel mask, ignoring slope '
                     'coefficient.')
            if area_exponent is not None:
                warn('Channel mask and area '
                     'exponent supplied. Defaulting '
                     'to channel mask, ignoring area '
                     'exponent.')
            if slope_exponent is not None:
                warn('Channel mask and slope '
                     'exponent supplied. Defaulting '
                     'to channel mask, ignoring slope '
                     'exponent.')
            if channelization_threshold is not None:
                warn('Channel mask and channelization '
                     'threshold supplied. Defaulting '
                     'to channel mask, ignoring '
                     'threshold.')
            if grid.number_of_nodes != len(channel__mask):
                raise ValueError('Length of channel mask is not equal to '
                                 'number of grid nodes')
            if 'channel__mask' in grid.at_node:
                warn("Existing channel__mask grid field was overwritten.")

            grid.at_node['channel__mask'] = channel__mask

        if channel__mask is None:
            if area_coefficient is None:
                raise FieldError('No channel mask and no area '
                                 'coefficient supplied. Either '
                                 'a channel mask or all 5 threshold '
                                 'parameters are needed.')
            if slope_coefficient is None:
                raise FieldError('No channel mask and no slope '
                                 'coefficient supplied. Either '
                                 'a channel mask or all 5 threshold '
                                 'parameters are needed.')
            if area_exponent is None:
                raise FieldError('No channel mask and no area '
                                 'exponent supplied. Either '
                                 'a channel mask or all 5 threshold '
                                 'parameters are needed.')
            if slope_exponent is None:
                raise FieldError('No channel mask and no slope '
                                 'exponent supplied. Either '
                                 'a channel mask or all 5 threshold '
                                 'parameters are needed.')
            if channelization_threshold is None:
                raise FieldError('No channel mask and no channelization '
                                 'threshold supplied. Either '
                                 'a channel mask or all 5 threshold '
                                 'parameters are needed.')
            channel__mask = (area_coefficient * \
                np.power(grid.at_node['drainage_area'], area_exponent) \
                * slope_coefficient * \
                np.power(grid.at_node['topographic__steepest_slope'], \
                slope_exponent)) > channelization_threshold
            grid.at_node['channel__mask'] = channel__mask

        required = ('flow__receiver_node', 'flow__link_to_receiver_node',
                    'topographic__steepest_slope')
        for name in required:
            if name not in grid.at_node:
                raise FieldError(
                    '{name}: missing required field'.format(name=name))

        # Store grid
        self._grid = grid

        # for this component to work with Cython acceleration,
        # the channel_network must be uint8, not bool...
        self.channel_network = (
            grid.at_node['channel__mask'].view(dtype=np.uint8))

        # Flow receivers
        self.flow_receivers = grid.at_node['flow__receiver_node']

        # Links to receiver nodes
        self.stack_links = grid.at_node['flow__link_to_receiver_node']

        # Distance to channel
        try:
            self.distance_to_channel = grid.at_node[
                'surface_to_channel__minimum_distance']
        except KeyError:
            self.distance_to_channel = grid.add_zeros(
                'surface_to_channel__minimum_distance', at='node', dtype=float)

    def calc_drainage_density(self):
        """Calculate drainage density.

        Returns
        -------
        float
            The drainage density.
        """

        # ^there is no 'run_one_step' method b/c this is a tool, not a model.
        """Calculate distance to channel and drainage density, after
        Tucker et al., 2001.

        Returns
        -------
        landscape_drainage_density : float (1/m)
            Drainage density over the model domain.
        """
        from .cfuncs import _calc_dists_to_channel
        _calc_dists_to_channel(self.channel_network,
                               self.flow_receivers,
                               self.grid.length_of_d8,
                               self.stack_links,
                               self.distance_to_channel,
                               self.grid.number_of_nodes)
        landscape_drainage_density = 1. / (2.0 * np.mean(self.grid.at_node[
            'surface_to_channel__minimum_distance'][self.grid.core_nodes]))
        # self.distance_to_channel))  # this is THE drainage density
        return landscape_drainage_density
